# Copyright 2012 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import datetime

from iso8601 import iso8601
from oslo_utils import timeutils
import webob.exc

from cinder.api.contrib import services
from cinder.api import extensions
from cinder import context
from cinder import db
from cinder import exception
from cinder import policy
from cinder import test
from cinder.tests.unit.api import fakes


fake_services_list = [
    {'binary': 'cinder-scheduler',
     'host': 'host1',
     'availability_zone': 'cinder',
     'id': 1,
     'disabled': True,
     'updated_at': datetime.datetime(2012, 10, 29, 13, 42, 2),
     'created_at': datetime.datetime(2012, 9, 18, 2, 46, 27),
     'disabled_reason': 'test1',
     'modified_at': ''},
    {'binary': 'cinder-volume',
     'host': 'host1',
     'availability_zone': 'cinder',
     'id': 2,
     'disabled': True,
     'updated_at': datetime.datetime(2012, 10, 29, 13, 42, 5),
     'created_at': datetime.datetime(2012, 9, 18, 2, 46, 27),
     'disabled_reason': 'test2',
     'modified_at': ''},
    {'binary': 'cinder-scheduler',
     'host': 'host2',
     'availability_zone': 'cinder',
     'id': 3,
     'disabled': False,
     'updated_at': datetime.datetime(2012, 9, 19, 6, 55, 34),
     'created_at': datetime.datetime(2012, 9, 18, 2, 46, 28),
     'disabled_reason': '',
     'modified_at': ''},
    {'binary': 'cinder-volume',
     'host': 'host2',
     'availability_zone': 'cinder',
     'id': 4,
     'disabled': True,
     'updated_at': datetime.datetime(2012, 9, 18, 8, 3, 38),
     'created_at': datetime.datetime(2012, 9, 18, 2, 46, 28),
     'disabled_reason': 'test4',
     'modified_at': ''},
    {'binary': 'cinder-volume',
     'host': 'host2',
     'availability_zone': 'cinder',
     'id': 5,
     'disabled': True,
     'updated_at': datetime.datetime(2012, 9, 18, 8, 3, 38),
     'created_at': datetime.datetime(2012, 9, 18, 2, 46, 28),
     'disabled_reason': 'test5',
     'modified_at': datetime.datetime(2012, 10, 29, 13, 42, 5)},
    {'binary': 'cinder-volume',
     'host': 'host2',
     'availability_zone': 'cinder',
     'id': 6,
     'disabled': False,
     'updated_at': datetime.datetime(2012, 9, 18, 8, 3, 38),
     'created_at': datetime.datetime(2012, 9, 18, 2, 46, 28),
     'disabled_reason': '',
     'modified_at': datetime.datetime(2012, 9, 18, 8, 1, 38)},
    {'binary': 'cinder-scheduler',
     'host': 'host2',
     'availability_zone': 'cinder',
     'id': 6,
     'disabled': False,
     'updated_at': None,
     'created_at': datetime.datetime(2012, 9, 18, 2, 46, 28),
     'disabled_reason': '',
     'modified_at': None},
]


class FakeRequest(object):
    environ = {"cinder.context": context.get_admin_context()}
    GET = {}


# NOTE(uni): deprecating service request key, binary takes precedence
# Still keeping service key here for API compatibility sake.
class FakeRequestWithService(object):
    environ = {"cinder.context": context.get_admin_context()}
    GET = {"service": "cinder-volume"}


class FakeRequestWithBinary(object):
    environ = {"cinder.context": context.get_admin_context()}
    GET = {"binary": "cinder-volume"}


class FakeRequestWithHost(object):
    environ = {"cinder.context": context.get_admin_context()}
    GET = {"host": "host1"}


# NOTE(uni): deprecating service request key, binary takes precedence
# Still keeping service key here for API compatibility sake.
class FakeRequestWithHostService(object):
    environ = {"cinder.context": context.get_admin_context()}
    GET = {"host": "host1", "service": "cinder-volume"}


class FakeRequestWithHostBinary(object):
    environ = {"cinder.context": context.get_admin_context()}
    GET = {"host": "host1", "binary": "cinder-volume"}


def fake_service_get_all(context, filters=None):
    filters = filters or {}
    host = filters.get('host')
    binary = filters.get('binary')
    return [s for s in fake_services_list
            if (not host or s['host'] == host or
                s['host'].startswith(host + '@'))
            and (not binary or s['binary'] == binary)]


def fake_service_get_by_host_binary(context, host, binary):
    for service in fake_services_list:
        if service['host'] == host and service['binary'] == binary:
            return service
    return None


def fake_service_get_by_id(value):
    for service in fake_services_list:
        if service['id'] == value:
            return service
    return None


def fake_service_update(context, service_id, values):
    service = fake_service_get_by_id(service_id)
    if service is None:
        raise exception.ServiceNotFound(service_id=service_id)
    else:
        {'host': 'host1', 'service': 'cinder-volume',
         'disabled': values['disabled']}


def fake_policy_enforce(context, action, target):
    pass


def fake_utcnow(with_timezone=False):
    tzinfo = iso8601.Utc() if with_timezone else None
    return datetime.datetime(2012, 10, 29, 13, 42, 11, tzinfo=tzinfo)


class ServicesTest(test.TestCase):

    def setUp(self):
        super(ServicesTest, self).setUp()

        self.stubs.Set(db, "service_get_all", fake_service_get_all)
        self.stubs.Set(timeutils, "utcnow", fake_utcnow)
        self.stubs.Set(db, "service_get_by_args",
                       fake_service_get_by_host_binary)
        self.stubs.Set(db, "service_update", fake_service_update)
        self.stubs.Set(policy, "enforce", fake_policy_enforce)

        self.context = context.get_admin_context()
        self.ext_mgr = extensions.ExtensionManager()
        self.ext_mgr.extensions = {}
        self.controller = services.ServiceController(self.ext_mgr)

    def test_services_list(self):
        req = FakeRequest()
        res_dict = self.controller.index(req)

        response = {'services': [{'binary': 'cinder-scheduler',
                                  'host': 'host1', 'zone': 'cinder',
                                  'status': 'disabled', 'state': 'up',
                                  'updated_at': datetime.datetime(
                                      2012, 10, 29, 13, 42, 2)},
                                 {'binary': 'cinder-volume',
                                  'host': 'host1', 'zone': 'cinder',
                                  'status': 'disabled', 'state': 'up',
                                  'updated_at': datetime.datetime(
                                      2012, 10, 29, 13, 42, 5)},
                                 {'binary': 'cinder-scheduler',
                                  'host': 'host2',
                                  'zone': 'cinder',
                                  'status': 'enabled', 'state': 'down',
                                  'updated_at': datetime.datetime(
                                      2012, 9, 19, 6, 55, 34)},
                                 {'binary': 'cinder-volume',
                                  'host': 'host2',
                                  'zone': 'cinder',
                                  'status': 'disabled', 'state': 'down',
                                  'updated_at': datetime.datetime(
                                      2012, 9, 18, 8, 3, 38)},
                                 {'binary': 'cinder-volume',
                                  'host': 'host2',
                                  'zone': 'cinder',
                                  'status': 'disabled', 'state': 'down',
                                  'updated_at': datetime.datetime(
                                      2012, 10, 29, 13, 42, 5)},
                                 {'binary': 'cinder-volume',
                                  'host': 'host2',
                                  'zone': 'cinder',
                                  'status': 'enabled', 'state': 'down',
                                  'updated_at': datetime.datetime(
                                      2012, 9, 18, 8, 3, 38)},
                                 {'binary': 'cinder-scheduler',
                                  'host': 'host2',
                                  'zone': 'cinder',
                                  'status': 'enabled', 'state': 'down',
                                  'updated_at': None},
                                 ]}
        self.assertEqual(response, res_dict)

    def test_services_detail(self):
        self.ext_mgr.extensions['os-extended-services'] = True
        self.controller = services.ServiceController(self.ext_mgr)
        req = FakeRequest()
        res_dict = self.controller.index(req)

        response = {'services': [{'binary': 'cinder-scheduler',
                                  'host': 'host1', 'zone': 'cinder',
                                  'status': 'disabled', 'state': 'up',
                                  'updated_at': datetime.datetime(
                                      2012, 10, 29, 13, 42, 2),
                                  'disabled_reason': 'test1'},
                                 {'binary': 'cinder-volume',
                                  'replication_status': None,
                                  'active_backend_id': None,
                                  'frozen': False,
                                  'host': 'host1', 'zone': 'cinder',
                                  'status': 'disabled', 'state': 'up',
                                  'updated_at': datetime.datetime(
                                      2012, 10, 29, 13, 42, 5),
                                  'disabled_reason': 'test2'},
                                 {'binary': 'cinder-scheduler',
                                  'host': 'host2',
                                  'zone': 'cinder',
                                  'status': 'enabled', 'state': 'down',
                                  'updated_at': datetime.datetime(
                                      2012, 9, 19, 6, 55, 34),
                                  'disabled_reason': ''},
                                 {'binary': 'cinder-volume',
                                  'replication_status': None,
                                  'active_backend_id': None,
                                  'frozen': False,
                                  'host': 'host2',
                                  'zone': 'cinder',
                                  'status': 'disabled', 'state': 'down',
                                  'updated_at': datetime.datetime(
                                      2012, 9, 18, 8, 3, 38),
                                  'disabled_reason': 'test4'},
                                 {'binary': 'cinder-volume',
                                  'replication_status': None,
                                  'active_backend_id': None,
                                  'frozen': False,
                                  'host': 'host2',
                                  'zone': 'cinder',
                                  'status': 'disabled', 'state': 'down',
                                  'updated_at': datetime.datetime(
                                      2012, 10, 29, 13, 42, 5),
                                  'disabled_reason': 'test5'},
                                 {'binary': 'cinder-volume',
                                  'replication_status': None,
                                  'active_backend_id': None,
                                  'frozen': False,
                                  'host': 'host2',
                                  'zone': 'cinder',
                                  'status': 'enabled', 'state': 'down',
                                  'updated_at': datetime.datetime(
                                      2012, 9, 18, 8, 3, 38),
                                  'disabled_reason': ''},
                                 {'binary': 'cinder-scheduler',
                                  'host': 'host2',
                                  'zone': 'cinder',
                                  'status': 'enabled', 'state': 'down',
                                  'updated_at': None,
                                  'disabled_reason': ''},
                                 ]}
        self.assertEqual(response, res_dict)

    def test_services_list_with_host(self):
        req = FakeRequestWithHost()
        res_dict = self.controller.index(req)

        response = {'services': [
            {'binary': 'cinder-scheduler',
             'host': 'host1',
             'zone': 'cinder',
             'status': 'disabled', 'state': 'up',
             'updated_at': datetime.datetime(2012, 10,
                                             29, 13, 42, 2)},
            {'binary': 'cinder-volume',
             'host': 'host1',
             'zone': 'cinder',
             'status': 'disabled', 'state': 'up',
             'updated_at': datetime.datetime(2012, 10, 29,
                                             13, 42, 5)}]}
        self.assertEqual(response, res_dict)

    def test_services_detail_with_host(self):
        self.ext_mgr.extensions['os-extended-services'] = True
        self.controller = services.ServiceController(self.ext_mgr)
        req = FakeRequestWithHost()
        res_dict = self.controller.index(req)

        response = {'services': [
            {'binary': 'cinder-scheduler',
             'host': 'host1',
             'zone': 'cinder',
             'status': 'disabled', 'state': 'up',
             'updated_at': datetime.datetime(2012, 10,
                                             29, 13, 42, 2),
             'disabled_reason': 'test1'},
            {'binary': 'cinder-volume',
             'frozen': False,
             'replication_status': None,
             'active_backend_id': None,
             'host': 'host1',
             'zone': 'cinder',
             'status': 'disabled', 'state': 'up',
             'updated_at': datetime.datetime(2012, 10, 29,
                                             13, 42, 5),
             'disabled_reason': 'test2'}]}
        self.assertEqual(response, res_dict)

    def test_services_list_with_service(self):
        req = FakeRequestWithService()
        res_dict = self.controller.index(req)

        response = {'services': [
            {'binary': 'cinder-volume',
             'host': 'host1',
             'zone': 'cinder',
             'status': 'disabled',
             'state': 'up',
             'updated_at': datetime.datetime(2012, 10, 29,
                                             13, 42, 5)},
            {'binary': 'cinder-volume',
             'host': 'host2',
             'zone': 'cinder',
             'status': 'disabled',
             'state': 'down',
             'updated_at': datetime.datetime(2012, 9, 18,
                                             8, 3, 38)},
            {'binary': 'cinder-volume',
             'host': 'host2',
             'zone': 'cinder',
             'status': 'disabled',
             'state': 'down',
             'updated_at': datetime.datetime(2012, 10, 29,
                                             13, 42, 5)},
            {'binary': 'cinder-volume',
             'host': 'host2',
             'zone': 'cinder',
             'status': 'enabled',
             'state': 'down',
             'updated_at': datetime.datetime(2012, 9, 18,
                                             8, 3, 38)}]}
        self.assertEqual(response, res_dict)

    def test_services_detail_with_service(self):
        self.ext_mgr.extensions['os-extended-services'] = True
        self.controller = services.ServiceController(self.ext_mgr)
        req = FakeRequestWithService()
        res_dict = self.controller.index(req)

        response = {'services': [
            {'binary': 'cinder-volume',
             'replication_status': None,
             'active_backend_id': None,
             'frozen': False,
             'host': 'host1',
             'zone': 'cinder',
             'status': 'disabled',
             'state': 'up',
             'updated_at': datetime.datetime(2012, 10, 29,
                                             13, 42, 5),
             'disabled_reason': 'test2'},
            {'binary': 'cinder-volume',
             'replication_status': None,
             'active_backend_id': None,
             'frozen': False,
             'host': 'host2',
             'zone': 'cinder',
             'status': 'disabled',
             'state': 'down',
             'updated_at': datetime.datetime(2012, 9, 18,
                                             8, 3, 38),
             'disabled_reason': 'test4'},
            {'binary': 'cinder-volume',
             'replication_status': None,
             'active_backend_id': None,
             'frozen': False,
             'host': 'host2',
             'zone': 'cinder',
             'status': 'disabled',
             'state': 'down',
             'updated_at': datetime.datetime(2012, 10, 29,
                                             13, 42, 5),
             'disabled_reason': 'test5'},
            {'binary': 'cinder-volume',
             'replication_status': None,
             'active_backend_id': None,
             'frozen': False,
             'host': 'host2',
             'zone': 'cinder',
             'status': 'enabled',
             'state': 'down',
             'updated_at': datetime.datetime(2012, 9, 18,
                                             8, 3, 38),
             'disabled_reason': ''}]}
        self.assertEqual(response, res_dict)

    def test_services_list_with_binary(self):
        req = FakeRequestWithBinary()
        res_dict = self.controller.index(req)

        response = {'services': [
            {'binary': 'cinder-volume',
             'host': 'host1',
             'zone': 'cinder',
             'status': 'disabled',
             'state': 'up',
             'updated_at': datetime.datetime(2012, 10, 29,
                                             13, 42, 5)},
            {'binary': 'cinder-volume',
             'host': 'host2',
             'zone': 'cinder',
             'status': 'disabled',
             'state': 'down',
             'updated_at': datetime.datetime(2012, 9, 18,
                                             8, 3, 38)},
            {'binary': 'cinder-volume',
             'host': 'host2',
             'zone': 'cinder',
             'status': 'disabled',
             'state': 'down',
             'updated_at': datetime.datetime(2012, 10, 29,
                                             13, 42, 5)},
            {'binary': 'cinder-volume',
             'host': 'host2',
             'zone': 'cinder',
             'status': 'enabled',
             'state': 'down',
             'updated_at': datetime.datetime(2012, 9, 18,
                                             8, 3, 38)}]}
        self.assertEqual(response, res_dict)

    def test_services_detail_with_binary(self):
        self.ext_mgr.extensions['os-extended-services'] = True
        self.controller = services.ServiceController(self.ext_mgr)
        req = FakeRequestWithBinary()
        res_dict = self.controller.index(req)

        response = {'services': [
            {'binary': 'cinder-volume',
             'replication_status': None,
             'active_backend_id': None,
             'host': 'host1',
             'zone': 'cinder',
             'status': 'disabled',
             'state': 'up',
             'frozen': False,
             'updated_at': datetime.datetime(2012, 10, 29,
                                             13, 42, 5),
             'disabled_reason': 'test2'},
            {'binary': 'cinder-volume',
             'replication_status': None,
             'active_backend_id': None,
             'host': 'host2',
             'zone': 'cinder',
             'status': 'disabled',
             'state': 'down',
             'frozen': False,
             'updated_at': datetime.datetime(2012, 9, 18,
                                             8, 3, 38),
             'disabled_reason': 'test4'},
            {'binary': 'cinder-volume',
             'replication_status': None,
             'active_backend_id': None,
             'host': 'host2',
             'zone': 'cinder',
             'status': 'disabled',
             'state': 'down',
             'frozen': False,
             'updated_at': datetime.datetime(2012, 10, 29,
                                             13, 42, 5),
             'disabled_reason': 'test5'},
            {'binary': 'cinder-volume',
             'replication_status': None,
             'active_backend_id': None,
             'host': 'host2',
             'zone': 'cinder',
             'status': 'enabled',
             'state': 'down',
             'frozen': False,
             'updated_at': datetime.datetime(2012, 9, 18,
                                             8, 3, 38),
             'disabled_reason': ''}]}
        self.assertEqual(response, res_dict)

    def test_services_list_with_host_service(self):
        req = FakeRequestWithHostService()
        res_dict = self.controller.index(req)

        response = {'services': [
            {'binary': 'cinder-volume',
             'host': 'host1',
             'zone': 'cinder',
             'status': 'disabled',
             'state': 'up',
             'updated_at': datetime.datetime(2012, 10, 29,
                                             13, 42, 5)}]}
        self.assertEqual(response, res_dict)

    def test_services_detail_with_host_service(self):
        self.ext_mgr.extensions['os-extended-services'] = True
        self.controller = services.ServiceController(self.ext_mgr)
        req = FakeRequestWithHostService()
        res_dict = self.controller.index(req)

        response = {'services': [
            {'binary': 'cinder-volume',
             'replication_status': None,
             'active_backend_id': None,
             'host': 'host1',
             'zone': 'cinder',
             'status': 'disabled',
             'state': 'up',
             'updated_at': datetime.datetime(2012, 10, 29,
                                             13, 42, 5),
             'disabled_reason': 'test2',
             'frozen': False}]}
        self.assertEqual(response, res_dict)

    def test_services_list_with_host_binary(self):
        req = FakeRequestWithHostBinary()
        res_dict = self.controller.index(req)

        response = {'services': [
            {'binary': 'cinder-volume',
             'host': 'host1',
             'zone': 'cinder',
             'status': 'disabled',
             'state': 'up',
             'updated_at': datetime.datetime(2012, 10, 29,
                                             13, 42, 5)}]}
        self.assertEqual(response, res_dict)

    def test_services_detail_with_host_binary(self):
        self.ext_mgr.extensions['os-extended-services'] = True
        self.controller = services.ServiceController(self.ext_mgr)
        req = FakeRequestWithHostBinary()
        res_dict = self.controller.index(req)

        response = {'services': [
            {'binary': 'cinder-volume',
             'replication_status': None,
             'active_backend_id': None,
             'frozen': False,
             'host': 'host1',
             'zone': 'cinder',
             'status': 'disabled',
             'state': 'up',
             'updated_at': datetime.datetime(2012, 10, 29,
                                             13, 42, 5),
             'disabled_reason': 'test2'}]}
        self.assertEqual(response, res_dict)

    def test_services_enable_with_service_key(self):
        body = {'host': 'host1', 'service': 'cinder-volume'}
        req = fakes.HTTPRequest.blank('/v2/fake/os-services/enable')
        res_dict = self.controller.update(req, "enable", body)

        self.assertEqual('enabled', res_dict['status'])

    def test_services_enable_with_binary_key(self):
        body = {'host': 'host1', 'binary': 'cinder-volume'}
        req = fakes.HTTPRequest.blank('/v2/fake/os-services/enable')
        res_dict = self.controller.update(req, "enable", body)

        self.assertEqual('enabled', res_dict['status'])

    def test_services_disable_with_service_key(self):
        req = fakes.HTTPRequest.blank('/v2/fake/os-services/disable')
        body = {'host': 'host1', 'service': 'cinder-volume'}
        res_dict = self.controller.update(req, "disable", body)

        self.assertEqual('disabled', res_dict['status'])

    def test_services_disable_with_binary_key(self):
        req = fakes.HTTPRequest.blank('/v2/fake/os-services/disable')
        body = {'host': 'host1', 'binary': 'cinder-volume'}
        res_dict = self.controller.update(req, "disable", body)

        self.assertEqual('disabled', res_dict['status'])

    def test_services_disable_log_reason(self):
        self.ext_mgr.extensions['os-extended-services'] = True
        self.controller = services.ServiceController(self.ext_mgr)
        req = (
            fakes.HTTPRequest.blank('v1/fake/os-services/disable-log-reason'))
        body = {'host': 'host1',
                'binary': 'cinder-scheduler',
                'disabled_reason': 'test-reason',
                }
        res_dict = self.controller.update(req, "disable-log-reason", body)

        self.assertEqual('disabled', res_dict['status'])
        self.assertEqual('test-reason', res_dict['disabled_reason'])

    def test_services_disable_log_reason_none(self):
        self.ext_mgr.extensions['os-extended-services'] = True
        self.controller = services.ServiceController(self.ext_mgr)
        req = (
            fakes.HTTPRequest.blank('v1/fake/os-services/disable-log-reason'))
        body = {'host': 'host1',
                'binary': 'cinder-scheduler',
                'disabled_reason': None,
                }
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.update,
                          req, "disable-log-reason", body)

    def test_invalid_reason_field(self):
        reason = ' '
        self.assertFalse(self.controller._is_valid_as_reason(reason))
        reason = 'a' * 256
        self.assertFalse(self.controller._is_valid_as_reason(reason))
        reason = 'it\'s a valid reason.'
        self.assertTrue(self.controller._is_valid_as_reason(reason))
        reason = None
        self.assertFalse(self.controller._is_valid_as_reason(reason))



# Copyright (C) 2014, 2015, Hitachi, Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import os
import re
import shlex
import threading
import time

from oslo_concurrency import processutils as putils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils import excutils
from oslo_utils import units
import six

from cinder import exception
from cinder.i18n import _LE, _LI, _LW
from cinder import utils
from cinder.volume.drivers.hitachi import hbsd_basiclib as basic_lib

GETSTORAGEARRAY_ONCE = 100
MAX_SNAPSHOT_COUNT = 1021
SNAP_LAST_PATH_SSB = '0xB958,0x020A'
HOST_IO_SSB = '0xB958,0x0233'
INVALID_LUN_SSB = '0x2E20,0x0000'
INTERCEPT_LDEV_SSB = '0x2E22,0x0001'
HOSTGROUP_INSTALLED = '0xB956,0x3173'
RESOURCE_LOCKED = 'SSB=0x2E11,0x2205'

LDEV_STATUS_WAITTIME = 120
LUN_DELETE_WAITTIME = basic_lib.DEFAULT_PROCESS_WAITTIME
LUN_DELETE_INTERVAL = 3
EXEC_MAX_WAITTIME = 30
EXEC_RETRY_INTERVAL = 5
HORCM_WAITTIME = 1
PAIR_TYPE = ('HORC', 'MRCF', 'QS')
PERMITTED_TYPE = ('CVS', 'HDP', 'HDT')

RAIDCOM_LOCK_FILE = basic_lib.LOCK_DIR + 'raidcom_'
HORCMGR_LOCK_FILE = basic_lib.LOCK_DIR + 'horcmgr_'
RESOURCE_LOCK_FILE = basic_lib.LOCK_DIR + 'raidcom_resource_'

STATUS_TABLE = {
    'SMPL': basic_lib.SMPL,
    'COPY': basic_lib.COPY,
    'RCPY': basic_lib.COPY,
    'PAIR': basic_lib.PAIR,
    'PFUL': basic_lib.PAIR,
    'PSUS': basic_lib.PSUS,
    'PFUS': basic_lib.PSUS,
    'SSUS': basic_lib.PSUS,
    'PSUE': basic_lib.PSUE,
}
NOT_SET = '-'
HORCM_RUNNING = 1
COPY_GROUP = basic_lib.NAME_PREFIX + '%s%s%03X%d'
SNAP_NAME = basic_lib.NAME_PREFIX + 'snap'
LDEV_NAME = basic_lib.NAME_PREFIX + 'ldev-%d-%d'
MAX_MUNS = 3

EX_ENAUTH = 202
EX_ENOOBJ = 205
EX_CMDRJE = 221
EX_CMDIOE = 237
EX_INVCMD = 240
EX_INVMOD = 241
EX_ENODEV = 246
EX_ENOENT = 247
EX_OPTINV = 248
EX_ATTDBG = 250
EX_ATTHOR = 251
EX_COMERR = 255
EX_UNKOWN = -1

NO_SUCH_DEVICE = (EX_ENODEV, EX_ENOENT)

COMMAND_IO_TO_RAID = (EX_CMDRJE, EX_CMDIOE, EX_INVCMD, EX_INVMOD, EX_OPTINV)

HORCM_ERROR = (EX_ATTDBG, EX_ATTHOR, EX_COMERR)

MAX_HOSTGROUPS = 254
MAX_HLUN = 2047

DEFAULT_PORT_BASE = 31000

LOG = logging.getLogger(__name__)

volume_opts = [
    cfg.StrOpt('hitachi_horcm_numbers',
               default='200,201',
               help='Instance numbers for HORCM'),
    cfg.StrOpt('hitachi_horcm_user',
               help='Username of storage system for HORCM'),
    cfg.StrOpt('hitachi_horcm_password',
               help='Password of storage system for HORCM',
               secret=True),
    cfg.BoolOpt('hitachi_horcm_add_conf',
                default=True,
                help='Add to HORCM configuration'),
    cfg.IntOpt('hitachi_horcm_resource_lock_timeout',
               default=600,
               help='Timeout until a resource lock is released, in seconds. '
                    'The value must be between 0 and 7200.'),
]

CONF = cfg.CONF
CONF.register_opts(volume_opts)


def horcm_synchronized(function):
    @functools.wraps(function)
    def wrapper(*args, **kargs):
        if len(args) == 1:
            inst = args[0].conf.hitachi_horcm_numbers[0]
            raidcom_obj_lock = args[0].raidcom_lock
        else:
            inst = args[1]
            raidcom_obj_lock = args[0].raidcom_pair_lock
        raidcom_lock_file = '%s%d' % (RAIDCOM_LOCK_FILE, inst)
        lock = basic_lib.get_process_lock(raidcom_lock_file)
        with raidcom_obj_lock, lock:
            return function(*args, **kargs)
    return wrapper


def storage_synchronized(function):
    @functools.wraps(function)
    def wrapper(*args, **kargs):
        serial = args[0].conf.hitachi_serial_number
        resource_lock = args[0].resource_lock
        resource_lock_file = '%s%s' % (RESOURCE_LOCK_FILE, serial)
        lock = basic_lib.get_process_lock(resource_lock_file)
        with resource_lock, lock:
            return function(*args, **kargs)
    return wrapper


class HBSDHORCM(basic_lib.HBSDBasicLib):

    def __init__(self, conf):
        super(HBSDHORCM, self).__init__(conf=conf)

        self.copy_groups = [None] * MAX_MUNS
        self.raidcom_lock = threading.Lock()
        self.raidcom_pair_lock = threading.Lock()
        self.horcmgr_lock = threading.Lock()
        self.horcmgr_flock = None
        self.resource_lock = threading.Lock()

    def check_param(self):
        numbers = self.conf.hitachi_horcm_numbers.split(',')
        if len(numbers) != 2:
            msg = basic_lib.output_err(601, param='hitachi_horcm_numbers')
            raise exception.HBSDError(message=msg)
        for i in numbers:
            if not i.isdigit():
                msg = basic_lib.output_err(601, param='hitachi_horcm_numbers')
                raise exception.HBSDError(message=msg)
        self.conf.hitachi_horcm_numbers = [int(num) for num in numbers]
        inst = self.conf.hitachi_horcm_numbers[0]
        pair_inst = self.conf.hitachi_horcm_numbers[1]
        if inst == pair_inst:
            msg = basic_lib.output_err(601, param='hitachi_horcm_numbers')
            raise exception.HBSDError(message=msg)
        for param in ('hitachi_horcm_user', 'hitachi_horcm_password'):
            if not getattr(self.conf, param):
                msg = basic_lib.output_err(601, param=param)
                raise exception.HBSDError(message=msg)
        if self.conf.hitachi_thin_pool_id == self.conf.hitachi_pool_id:
            msg = basic_lib.output_err(601, param='hitachi_thin_pool_id')
            raise exception.HBSDError(message=msg)
        resource_lock_timeout = self.conf.hitachi_horcm_resource_lock_timeout
        if not ((resource_lock_timeout >= 0) and
                (resource_lock_timeout <= 7200)):
            msg = basic_lib.output_err(
                601, param='hitachi_horcm_resource_lock_timeout')
            raise exception.HBSDError(message=msg)
        for opt in volume_opts:
            getattr(self.conf, opt.name)

    def set_copy_groups(self, host_ip):
        serial = self.conf.hitachi_serial_number
        inst = self.conf.hitachi_horcm_numbers[1]

        for mun in range(MAX_MUNS):
            copy_group = COPY_GROUP % (host_ip, serial, inst, mun)
            self.copy_groups[mun] = copy_group

    def set_pair_flock(self):
        inst = self.conf.hitachi_horcm_numbers[1]
        name = '%s%d' % (HORCMGR_LOCK_FILE, inst)
        self.horcmgr_flock = basic_lib.FileLock(name, self.horcmgr_lock)
        return self.horcmgr_flock

    def check_horcm(self, inst):
        args = 'HORCMINST=%d horcmgr -check' % inst
        ret, _stdout, _stderr = self.exec_command('env', args=args,
                                                  printflag=False)
        return ret

    def shutdown_horcm(self, inst):
        ret, stdout, stderr = self.exec_command(
            'horcmshutdown.sh', args=six.text_type(inst), printflag=False)
        return ret

    def start_horcm(self, inst):
        return self.exec_command('horcmstart.sh', args=six.text_type(inst),
                                 printflag=False)

    def _wait_for_horcm_shutdown(self, inst):
        if self.check_horcm(inst) != HORCM_RUNNING:
            raise loopingcall.LoopingCallDone()

        if self.shutdown_horcm(inst):
            LOG.error(_LE("Failed to shutdown horcm."))
            raise loopingcall.LoopingCallDone()

    @horcm_synchronized
    def restart_horcm(self, inst=None):
        if inst is None:
            inst = self.conf.hitachi_horcm_numbers[0]

        loop = loopingcall.FixedIntervalLoopingCall(
            self._wait_for_horcm_shutdown, inst)

        loop.start(interval=HORCM_WAITTIME).wait()

        ret, stdout, stderr = self.start_horcm(inst)
        if ret:
            msg = basic_lib.output_err(
                600, cmd='horcmstart.sh', ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

    def restart_pair_horcm(self):
        inst = self.conf.hitachi_horcm_numbers[1]
        self.restart_horcm(inst=inst)

    def setup_horcmgr(self, host_ip):
        pair_inst = self.conf.hitachi_horcm_numbers[1]
        self.set_copy_groups(host_ip)
        if self.conf.hitachi_horcm_add_conf:
            self.create_horcmconf()
            self.create_horcmconf(inst=pair_inst)
        self.restart_horcm()
        with self.horcmgr_flock:
            self.restart_pair_horcm()
        ret, stdout, stderr = self.comm_login()
        if ret:
            msg = basic_lib.output_err(
                600, cmd='raidcom -login', ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

    def _wait_for_exec_horcm(self, cmd, args, printflag, start):
        if cmd == 'raidcom':
            serial = self.conf.hitachi_serial_number
            inst = self.conf.hitachi_horcm_numbers[0]
            raidcom_obj_lock = self.raidcom_lock
            args = '%s -s %s -I%d' % (args, serial, inst)
        else:
            inst = self.conf.hitachi_horcm_numbers[1]
            raidcom_obj_lock = self.raidcom_pair_lock
            args = '%s -ISI%d' % (args, inst)
        user = self.conf.hitachi_horcm_user
        passwd = self.conf.hitachi_horcm_password
        raidcom_lock_file = '%s%d' % (RAIDCOM_LOCK_FILE, inst)
        lock = basic_lib.get_process_lock(raidcom_lock_file)

        with raidcom_obj_lock, lock:
            ret, stdout, stderr = self.exec_command(cmd, args=args,
                                                    printflag=printflag)

        # The resource group may be locked by other software.
        # Therefore, wait until the lock is released.
        if (RESOURCE_LOCKED in stderr and
            (time.time() - start <
             self.conf.hitachi_horcm_resource_lock_timeout)):
            return

        if not ret or ret <= 127:
            raise loopingcall.LoopingCallDone((ret, stdout, stderr))

        if time.time() - start >= EXEC_MAX_WAITTIME:
            LOG.error(_LE("horcm command timeout."))
            raise loopingcall.LoopingCallDone((ret, stdout, stderr))

        if (ret == EX_ENAUTH and
                not re.search("-login %s %s" % (user, passwd), args)):
            _ret, _stdout, _stderr = self.comm_login()
            if _ret:
                LOG.error(_LE("Failed to authenticate user."))
                raise loopingcall.LoopingCallDone((ret, stdout, stderr))

        elif ret in HORCM_ERROR:
            _ret = 0
            with raidcom_obj_lock, lock:
                if self.check_horcm(inst) != HORCM_RUNNING:
                    _ret, _stdout, _stderr = self.start_horcm(inst)
            if _ret and _ret != HORCM_RUNNING:
                LOG.error(_LE("Failed to start horcm."))
                raise loopingcall.LoopingCallDone((ret, stdout, stderr))

        elif ret not in COMMAND_IO_TO_RAID:
            LOG.error(_LE("Unexpected error occurs in horcm."))
            raise loopingcall.LoopingCallDone((ret, stdout, stderr))

    def exec_raidcom(self, cmd, args, printflag=True):
        loop = loopingcall.FixedIntervalLoopingCall(
            self._wait_for_exec_horcm, cmd, args, printflag, time.time())

        return loop.start(interval=EXEC_RETRY_INTERVAL).wait()

    def comm_login(self):
        rmi_user = self.conf.hitachi_horcm_user
        rmi_pass = self.conf.hitachi_horcm_password
        args = '-login %s %s' % (rmi_user, rmi_pass)
        return self.exec_raidcom('raidcom', args, printflag=False)

    def comm_reset_status(self):
        self.exec_raidcom('raidcom', 'reset command_status')

    def comm_get_status(self):
        return self.exec_raidcom('raidcom', 'get command_status')

    def get_command_error(self, stdout):
        lines = stdout.splitlines()
        line = shlex.split(lines[1])
        return int(line[3])

    def comm_get_ldev(self, ldev):
        opt = 'get ldev -ldev_id %s' % ldev
        ret, stdout, stderr = self.exec_raidcom('raidcom', opt,
                                                printflag=False)
        if ret:
            opt = 'raidcom %s' % opt
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)
        return stdout

    def add_used_hlun(self, port, gid, used_list):
        opt = 'get lun -port %s-%d' % (port, gid)
        ret, stdout, stderr = self.exec_raidcom('raidcom', opt,
                                                printflag=False)
        if ret:
            opt = 'raidcom %s' % opt
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)
        lines = stdout.splitlines()
        for line in lines[1:]:
            lun = int(shlex.split(line)[3])
            if lun not in used_list:
                used_list.append(lun)

    def get_unused_ldev(self, ldev_range):
        start = ldev_range[0]
        end = ldev_range[1]

        while start < end:
            if end - start + 1 > GETSTORAGEARRAY_ONCE:
                cnt = GETSTORAGEARRAY_ONCE
            else:
                cnt = end - start + 1
            opt = 'get ldev -ldev_id %d -cnt %d' % (start, cnt)
            ret, stdout, stderr = self.exec_raidcom('raidcom', opt,
                                                    printflag=False)
            if ret:
                opt = 'raidcom %s' % opt
                msg = basic_lib.output_err(
                    600, cmd=opt, ret=ret, out=stdout, err=stderr)
                raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

            lines = stdout.splitlines()
            ldev_num = None
            for line in lines:
                if re.match("LDEV :", line):
                    ldev_num = int(shlex.split(line)[2])
                    continue
                if re.match("VOL_TYPE : NOT DEFINED", line):
                    return ldev_num

            start += GETSTORAGEARRAY_ONCE
        else:
            msg = basic_lib.output_err(648, resource='LDEV')
            raise exception.HBSDError(message=msg)

    def get_hgname_gid(self, port, host_grp_name):
        opt = 'get host_grp -port %s -key host_grp' % port
        ret, stdout, stderr = self.exec_raidcom('raidcom', opt,
                                                printflag=False)
        if ret:
            opt = 'raidcom %s' % opt
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)
        lines = stdout.splitlines()
        for line in lines[1:]:
            line = shlex.split(line)
            if line[2] == host_grp_name:
                return int(line[1])
        return None

    def get_unused_gid(self, range, port):
        _min = range[0]
        _max = range[1]
        opt = 'get host_grp -port %s -key host_grp' % port
        ret, stdout, stderr = self.exec_raidcom('raidcom', opt,
                                                printflag=False)
        if ret:
            opt = 'raidcom %s' % opt
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

        lines = stdout.splitlines()
        free_gid = None
        for line in lines[_min + 1:]:
            line = shlex.split(line)
            if int(line[1]) > _max:
                break
            if line[2] == '-':
                free_gid = int(line[1])
                break
        if free_gid is None:
            msg = basic_lib.output_err(648, resource='GID')
            raise exception.HBSDError(message=msg)
        return free_gid

    def comm_set_target_wwns(self, target_ports):
        opt = 'get port'
        ret, stdout, stderr = self.exec_raidcom('raidcom', opt,
                                                printflag=False)
        if ret:
            opt = 'raidcom %s' % opt
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

        target_wwns = {}
        lines = stdout.splitlines()
        for line in lines[1:]:
            line = shlex.split(line)
            port = line[0][:5]
            if target_ports and port not in target_ports:
                continue

            target_wwns[port] = line[10]
        LOG.debug('target wwns: %s', target_wwns)
        return target_wwns

    def comm_get_hbawwn(self, hostgroups, wwns, port, is_detected):
        opt = 'get host_grp -port %s' % port
        ret, stdout, stderr = self.exec_raidcom('raidcom', opt,
                                                printflag=False)
        if ret:
            opt = 'raidcom %s' % opt
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

        lines = stdout.splitlines()
        found_wwns = 0
        for line in lines[1:]:
            line = shlex.split(line)
            if not re.match(basic_lib.NAME_PREFIX, line[2]):
                continue
            gid = line[1]
            opt = 'get hba_wwn -port %s-%s' % (port, gid)
            ret, stdout, stderr = self.exec_raidcom(
                'raidcom', opt, printflag=False)
            if ret:
                opt = 'raidcom %s' % opt
                msg = basic_lib.output_err(
                    600, cmd=opt, ret=ret, out=stdout, err=stderr)
                raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

            lines = stdout.splitlines()
            for line in lines[1:]:
                hba_info = shlex.split(line)

                if hba_info[3] in wwns:
                    hostgroups.append({'port': six.text_type(port),
                                       'gid': int(hba_info[1]),
                                       'initiator_wwn': hba_info[3],
                                       'detected': is_detected})
                    found_wwns += 1
                if len(wwns) == found_wwns:
                    break

            if len(wwns) == found_wwns:
                break

    def comm_chk_login_wwn(self, wwns, port):
        opt = 'get port -port %s' % port
        ret, stdout, stderr = self.exec_raidcom('raidcom', opt,
                                                printflag=False)

        if ret:
            opt = 'raidcom %s' % opt
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

        lines = stdout.splitlines()
        for line in lines[1:]:
            login_info = shlex.split(line)
            if login_info[1] in wwns:
                return True
        else:
            return False

    def comm_get_hostgroup_info(self, hgs, wwns, target_ports, login=True):
        security_ports = []
        hostgroups = []

        opt = 'get port'
        ret, stdout, stderr = self.exec_raidcom('raidcom', opt,
                                                printflag=False)
        if ret:
            opt = 'raidcom %s' % opt
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

        lines = stdout.splitlines()

        for line in lines[1:]:
            line = shlex.split(line)
            port = line[0][:5]
            if target_ports and port not in target_ports:
                continue
            security = True if line[7] == 'Y' else False

            is_detected = None
            if login:
                is_detected = self.comm_chk_login_wwn(wwns, port)

            if security:
                self.comm_get_hbawwn(hostgroups, wwns, port, is_detected)
                security_ports.append(port)

        for hostgroup in hostgroups:
            hgs.append(hostgroup)

        return security_ports

    def _get_lun(self, port, gid, ldev):
        lun = None

        opt = 'get lun -port %s-%d' % (port, gid)
        ret, stdout, stderr = self.exec_raidcom('raidcom', opt,
                                                printflag=False)
        if ret:
            opt = 'raidcom %s' % opt
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

        lines = stdout.splitlines()
        for line in lines[1:]:
            line = shlex.split(line)
            if line[5] == six.text_type(ldev):
                lun = int(line[3])
                break

        return lun

    def _wait_for_delete_lun(self, hostgroup, ldev, start):
        opt = 'delete lun -port %s-%d -ldev_id %d' % (hostgroup['port'],
                                                      hostgroup['gid'], ldev)
        ret, stdout, stderr = self.exec_raidcom('raidcom', opt)
        if not ret:
            raise loopingcall.LoopingCallDone()

        if (re.search('SSB=%s' % SNAP_LAST_PATH_SSB, stderr) and
                not self.comm_get_snapshot(ldev) or
                re.search('SSB=%s' % HOST_IO_SSB, stderr)):
            LOG.warning(basic_lib.set_msg(310, ldev=ldev, reason=stderr))

            if time.time() - start >= LUN_DELETE_WAITTIME:
                msg = basic_lib.output_err(
                    637, method='_wait_for_delete_lun',
                    timeout=LUN_DELETE_WAITTIME)
                raise exception.HBSDError(message=msg)
        else:
            opt = 'raidcom %s' % opt
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

    def comm_delete_lun_core(self, hostgroup, ldev):
        loop = loopingcall.FixedIntervalLoopingCall(
            self._wait_for_delete_lun, hostgroup, ldev, time.time())

        loop.start(interval=LUN_DELETE_INTERVAL).wait()

    def comm_delete_lun(self, hostgroups, ldev):
        deleted_hostgroups = []
        no_ldev_cnt = 0
        for hostgroup in hostgroups:
            port = hostgroup['port']
            gid = hostgroup['gid']
            is_deleted = False
            for deleted in deleted_hostgroups:
                if port == deleted['port'] and gid == deleted['gid']:
                    is_deleted = True
            if is_deleted:
                continue
            try:
                self.comm_delete_lun_core(hostgroup, ldev)
            except exception.HBSDCmdError as ex:
                no_ldev_cnt += 1
                if ex.ret == EX_ENOOBJ:
                    if no_ldev_cnt != len(hostgroups):
                        continue
                    raise exception.HBSDNotFound
                else:
                    raise
            deleted_hostgroups.append({'port': port, 'gid': gid})

    def _check_ldev_status(self, ldev, status):
        opt = ('get ldev -ldev_id %s -check_status %s -time %s' %
               (ldev, status, LDEV_STATUS_WAITTIME))
        ret, _stdout, _stderr = self.exec_raidcom('raidcom', opt)
        return ret

    # Don't remove a storage_syncronized decorator.
    # It is need to avoid comm_add_ldev() and comm_delete_ldev() are
    # executed concurrently.
    @storage_synchronized
    def comm_add_ldev(self, pool_id, ldev, capacity, is_vvol):
        emulation = 'OPEN-V'
        if is_vvol:
            opt = ('add ldev -pool snap -ldev_id %d '
                   '-capacity %dG -emulation %s'
                   % (ldev, capacity, emulation))
        else:
            opt = ('add ldev -pool %d -ldev_id %d '
                   '-capacity %dG -emulation %s'
                   % (pool_id, ldev, capacity, emulation))

        self.comm_reset_status()
        ret, stdout, stderr = self.exec_raidcom('raidcom', opt)
        if ret:
            if re.search('SSB=%s' % INTERCEPT_LDEV_SSB, stderr):
                raise exception.HBSDNotFound

            msg = basic_lib.output_err(
                600, cmd='raidcom %s' % opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

        if self._check_ldev_status(ldev, "NML"):
            msg = basic_lib.output_err(653, ldev=ldev)
            raise exception.HBSDError(message=msg)

    def comm_add_hostgrp(self, port, gid, host_grp_name):
        opt = 'add host_grp -port %s-%d -host_grp_name %s' % (port, gid,
                                                              host_grp_name)
        ret, stdout, stderr = self.exec_raidcom('raidcom', opt)
        if ret:
            if re.search('SSB=%s' % HOSTGROUP_INSTALLED, stderr):
                raise exception.HBSDNotFound

            msg = basic_lib.output_err(
                600, cmd='raidcom %s' % opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

    def comm_del_hostgrp(self, port, gid, host_grp_name):
        opt = 'delete host_grp -port %s-%d %s' % (port, gid, host_grp_name)
        ret, stdout, stderr = self.exec_raidcom('raidcom', opt)
        if ret:
            msg = basic_lib.output_err(
                600, cmd='raidcom %s' % opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

    def comm_add_hbawwn(self, port, gid, wwn):
        opt = 'add hba_wwn -port %s-%s -hba_wwn %s' % (port, gid, wwn)
        ret, stdout, stderr = self.exec_raidcom('raidcom', opt)
        if ret:
            msg = basic_lib.output_err(
                600, cmd='raidcom %s' % opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

    @storage_synchronized
    def comm_add_lun(self, unused_command, hostgroups, ldev, is_once=False):
        tmp_hostgroups = hostgroups[:]
        is_ok = False
        used_list = []
        lun = None
        old_lun = None

        for hostgroup in hostgroups:
            port = hostgroup['port']
            gid = hostgroup['gid']
            self.add_used_hlun(port, gid, used_list)
            lun = self._get_lun(port, gid, ldev)

            # When 'lun' or 'old_lun' is 0, it should be true.
            # So, it cannot remove 'is not None'.
            if lun is not None:
                if old_lun is not None and old_lun != lun:
                    msg = basic_lib.output_err(648, resource='LUN (HLUN)')
                    raise exception.HBSDError(message=msg)
                is_ok = True
                hostgroup['lun'] = lun
                tmp_hostgroups.remove(hostgroup)
                old_lun = lun

            if is_once:
                # When 'lun' is 0, it should be true.
                # So, it cannot remove 'is not None'.
                if lun is not None:
                    return
                elif len(used_list) < MAX_HLUN + 1:
                    break
                else:
                    tmp_hostgroups.remove(hostgroup)
                    if tmp_hostgroups:
                        used_list = []

        if not used_list:
            lun = 0
        elif lun is None:
            for i in range(MAX_HLUN + 1):
                if i not in used_list:
                    lun = i
                    break
            else:
                raise exception.HBSDNotFound

        opt = None
        ret = 0
        stdout = None
        stderr = None
        invalid_hgs_str = None

        for hostgroup in tmp_hostgroups:
            port = hostgroup['port']
            gid = hostgroup['gid']
            if not hostgroup['detected']:
                if invalid_hgs_str:
                    invalid_hgs_str = '%s, %s:%d' % (invalid_hgs_str,
                                                     port, gid)
                else:
                    invalid_hgs_str = '%s:%d' % (port, gid)
                continue
            opt = 'add lun -port %s-%d -ldev_id %d -lun_id %d' % (
                port, gid, ldev, lun)
            ret, stdout, stderr = self.exec_raidcom('raidcom', opt)
            if not ret:
                is_ok = True
                hostgroup['lun'] = lun
                if is_once:
                    break
            else:
                LOG.warning(basic_lib.set_msg(
                    314, ldev=ldev, lun=lun, port=port, id=gid))

        if not is_ok:
            if stderr:
                opt = 'raidcom %s' % opt
                msg = basic_lib.output_err(
                    600, cmd=opt, ret=ret, out=stdout, err=stderr)
                raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)
            else:
                msg = basic_lib.output_err(659, gid=invalid_hgs_str)
                raise exception.HBSDError(message=msg)

    # Don't remove a storage_syncronized decorator.
    # It is need to avoid comm_add_ldev() and comm_delete_ldev() are
    # executed concurrently.
    @storage_synchronized
    def comm_delete_ldev(self, ldev, is_vvol):
        ret = -1
        stdout = ""
        stderr = ""
        self.comm_reset_status()
        opt = 'delete ldev -ldev_id %d' % ldev
        ret, stdout, stderr = self.exec_raidcom('raidcom', opt)
        if ret:
            if re.search('SSB=%s' % INVALID_LUN_SSB, stderr):
                raise exception.HBSDNotFound

            msg = basic_lib.output_err(
                600, cmd='raidcom %s' % opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

        ret, stdout, stderr = self.comm_get_status()
        if ret or self.get_command_error(stdout):
            opt = 'raidcom %s' % opt
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

    def comm_extend_ldev(self, ldev, old_size, new_size):
        extend_size = new_size - old_size
        opt = 'extend ldev -ldev_id %d -capacity %dG' % (ldev, extend_size)
        ret, stdout, stderr = self.exec_raidcom('raidcom', opt)
        if ret:
            msg = basic_lib.output_err(
                600, cmd='raidcom %s' % opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

    def comm_get_dp_pool(self, pool_id):
        opt = 'get dp_pool'
        ret, stdout, stderr = self.exec_raidcom('raidcom', opt,
                                                printflag=False)
        if ret:
            opt = 'raidcom %s' % opt
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

        lines = stdout.splitlines()
        for line in lines[1:]:
            if int(shlex.split(line)[0]) == pool_id:
                free_gb = int(shlex.split(line)[3]) / 1024
                total_gb = int(shlex.split(line)[4]) / 1024
                return total_gb, free_gb

        msg = basic_lib.output_err(640, pool_id=pool_id)
        raise exception.HBSDError(message=msg)

    def comm_modify_ldev(self, ldev):
        args = 'modify ldev -ldev_id %d -status discard_zero_page' % ldev
        ret, stdout, stderr = self.exec_raidcom('raidcom', args)
        if ret:
            LOG.warning(basic_lib.set_msg(315, ldev=ldev, reason=stderr))

    def is_detected(self, port, wwn):
        return self.comm_chk_login_wwn([wwn], port)

    def discard_zero_page(self, ldev):
        try:
            self.comm_modify_ldev(ldev)
        except Exception as ex:
            LOG.warning(_LW('Failed to discard zero page: %s'), ex)

    def comm_add_snapshot(self, pvol, svol):
        pool = self.conf.hitachi_thin_pool_id
        copy_size = self.conf.hitachi_copy_speed
        args = ('add snapshot -ldev_id %d %d -pool %d '
                '-snapshot_name %s -copy_size %d'
                % (pvol, svol, pool, SNAP_NAME, copy_size))
        ret, stdout, stderr = self.exec_raidcom('raidcom', args)
        if ret:
            msg = basic_lib.output_err(
                600, cmd='raidcom %s' % args, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

    def comm_delete_snapshot(self, ldev):
        args = 'delete snapshot -ldev_id %d' % ldev
        ret, stdout, stderr = self.exec_raidcom('raidcom', args)
        if ret:
            msg = basic_lib.output_err(
                600, cmd='raidcom %s' % args, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

    def comm_modify_snapshot(self, ldev, op):
        args = ('modify snapshot -ldev_id %d -snapshot_data %s' % (ldev, op))
        ret, stdout, stderr = self.exec_raidcom('raidcom', args)
        if ret:
            msg = basic_lib.output_err(
                600, cmd='raidcom %s' % args, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

    def _wait_for_snap_status(self, pvol, svol, status, timeout, start):
        if (self.get_snap_pvol_status(pvol, svol) in status and
                self.get_snap_svol_status(svol) in status):
            raise loopingcall.LoopingCallDone()

        if time.time() - start >= timeout:
            msg = basic_lib.output_err(
                637, method='_wait_for_snap_status', timuout=timeout)
            raise exception.HBSDError(message=msg)

    def wait_snap(self, pvol, svol, status, timeout, interval):
        loop = loopingcall.FixedIntervalLoopingCall(
            self._wait_for_snap_status, pvol,
            svol, status, timeout, time.time())

        loop.start(interval=interval).wait()

    def comm_get_snapshot(self, ldev):
        args = 'get snapshot -ldev_id %d' % ldev
        ret, stdout, stderr = self.exec_raidcom('raidcom', args,
                                                printflag=False)
        if ret:
            opt = 'raidcom %s' % args
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)
        return stdout

    def check_snap_count(self, ldev):
        stdout = self.comm_get_snapshot(ldev)
        if not stdout:
            return
        lines = stdout.splitlines()
        if len(lines) >= MAX_SNAPSHOT_COUNT + 1:
            msg = basic_lib.output_err(
                615, copy_method=basic_lib.THIN, pvol=ldev)
            raise exception.HBSDBusy(message=msg)

    def get_snap_pvol_status(self, pvol, svol):
        stdout = self.comm_get_snapshot(pvol)
        if not stdout:
            return basic_lib.SMPL
        lines = stdout.splitlines()
        for line in lines[1:]:
            line = shlex.split(line)
            if int(line[6]) == svol:
                return STATUS_TABLE[line[2]]
        else:
            return basic_lib.SMPL

    def get_snap_svol_status(self, ldev):
        stdout = self.comm_get_snapshot(ldev)
        if not stdout:
            return basic_lib.SMPL
        lines = stdout.splitlines()
        line = shlex.split(lines[1])
        return STATUS_TABLE[line[2]]

    @horcm_synchronized
    def create_horcmconf(self, inst=None):
        if inst is None:
            inst = self.conf.hitachi_horcm_numbers[0]

        serial = self.conf.hitachi_serial_number
        filename = '/etc/horcm%d.conf' % inst

        port = DEFAULT_PORT_BASE + inst

        found = False

        if not os.path.exists(filename):
            file_str = """
HORCM_MON
#ip_address        service         poll(10ms)     timeout(10ms)
127.0.0.1 %16d               6000              3000
HORCM_CMD
""" % port
        else:
            file_str = utils.read_file_as_root(filename)

            lines = file_str.splitlines()
            for line in lines:
                if re.match(r'\\\\.\\CMD-%s:/dev/sd' % serial, line):
                    found = True
                    break

        if not found:
            insert_str = r'\\\\.\\CMD-%s:/dev/sd' % serial
            file_str = re.sub(r'(\n\bHORCM_CMD.*|^\bHORCM_CMD.*)',
                              r'\1\n%s\n' % insert_str, file_str)

            try:
                utils.execute('tee', filename, process_input=file_str,
                              run_as_root=True)
            except putils.ProcessExecutionError as ex:
                msg = basic_lib.output_err(
                    632, file=filename, ret=ex.exit_code, err=ex.stderr)
                raise exception.HBSDError(message=msg)

    def comm_get_copy_grp(self):
        ret, stdout, stderr = self.exec_raidcom('raidcom', 'get copy_grp',
                                                printflag=False)
        if ret:
            opt = 'raidcom get copy_grp'
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)
        return stdout

    def comm_add_copy_grp(self, copy_group, pvol_group, svol_group, mun):
        args = ('add copy_grp -copy_grp_name %s %s %s -mirror_id %d'
                % (copy_group, pvol_group, svol_group, mun))
        ret, stdout, stderr = self.exec_raidcom('raidcom', args,
                                                printflag=False)
        if ret:
            opt = 'raidcom %s' % args
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

    def comm_delete_copy_grp(self, copy_group):
        args = 'delete copy_grp -copy_grp_name %s' % copy_group
        ret, stdout, stderr = self.exec_raidcom('raidcom', args,
                                                printflag=False)
        if ret:
            opt = 'raidcom %s' % args
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

    def comm_get_device_grp(self, group_name):
        args = 'get device_grp -device_grp_name %s' % group_name
        ret, stdout, stderr = self.exec_raidcom('raidcom', args,
                                                printflag=False)
        if ret:
            opt = 'raidcom %s' % args
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)
        return stdout

    def comm_add_device_grp(self, group_name, ldev_name, ldev):
        args = ('add device_grp -device_grp_name %s %s -ldev_id %d'
                % (group_name, ldev_name, ldev))
        ret, stdout, stderr = self.exec_raidcom('raidcom', args,
                                                printflag=False)
        if ret:
            opt = 'raidcom %s' % args
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

    def comm_delete_device_grp(self, group_name, ldev):
        args = ('delete device_grp -device_grp_name %s -ldev_id %d'
                % (group_name, ldev))
        ret, stdout, stderr = self.exec_raidcom('raidcom', args,
                                                printflag=False)
        if ret:
            opt = 'raidcom %s' % args
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

    def comm_paircreate(self, copy_group, ldev_name):
        args = ('-g %s -d %s -split -fq quick -c %d -vl'
                % (copy_group, ldev_name, self.conf.hitachi_copy_speed))
        ret, stdout, stderr = self.exec_raidcom('paircreate', args)
        if ret:
            opt = 'paircreate %s' % args
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

    def comm_pairsplit(self, copy_group, ldev_name):
        args = '-g %s -d %s -S' % (copy_group, ldev_name)
        ret, stdout, stderr = self.exec_raidcom('pairsplit', args)
        if ret:
            opt = 'pairsplit %s' % args
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)

    def comm_pairevtwait(self, copy_group, ldev_name, check_svol):
        if not check_svol:
            option = '-nowait'
        else:
            option = '-nowaits'
        args = '-g %s -d %s %s' % (copy_group, ldev_name, option)
        ret, stdout, stderr = self.exec_raidcom('pairevtwait', args,
                                                printflag=False)
        if ret > 127:
            opt = 'pairevtwait %s' % args
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)
        return ret

    def comm_pairdisplay(self, copy_group, ldev_name=None):
        if not ldev_name:
            args = '-g %s -CLI' % copy_group
        else:
            args = '-g %s -d %s -CLI' % (copy_group, ldev_name)
        ret, stdout, stderr = self.exec_raidcom('pairdisplay', args,
                                                printflag=False)
        if ret and ret not in NO_SUCH_DEVICE:
            opt = 'pairdisplay %s' % args
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)
        return ret, stdout, stderr

    def check_copy_grp(self, copy_group):
        stdout = self.comm_get_copy_grp()
        lines = stdout.splitlines()
        count = 0
        for line in lines[1:]:
            line = shlex.split(line)
            if line[0] == copy_group:
                count += 1
                if count == 2:
                    break
        return count

    def check_device_grp(self, group_name, ldev, ldev_name=None):
        stdout = self.comm_get_device_grp(group_name)
        lines = stdout.splitlines()
        for line in lines[1:]:
            line = shlex.split(line)
            if int(line[2]) == ldev:
                if not ldev_name:
                    return True
                else:
                    return line[1] == ldev_name
        else:
            return False

    def is_smpl(self, copy_group, ldev_name):
        ret, stdout, stderr = self.comm_pairdisplay(copy_group,
                                                    ldev_name=ldev_name)
        if not stdout:
            return True

        lines = stdout.splitlines()
        for line in lines[1:]:
            line = shlex.split(line)
            if line[9] in [NOT_SET, 'SMPL']:
                return True
        else:
            return False

    def get_copy_groups(self):
        copy_groups = []
        stdout = self.comm_get_copy_grp()
        lines = stdout.splitlines()
        for line in lines[1:]:
            line = shlex.split(line)
            if line[0] in self.copy_groups and line[0] not in copy_groups:
                copy_groups.append(line[0])
        return copy_groups

    def get_matched_copy_group(self, pvol, svol, ldev_name):
        for copy_group in self.get_copy_groups():
            pvol_group = '%sP' % copy_group
            if self.check_device_grp(pvol_group, pvol, ldev_name=ldev_name):
                return copy_group
        else:
            return None

    def get_paired_info(self, ldev, only_flag=False):
        paired_info = {'pvol': None, 'svol': []}
        pvol = None
        is_svol = False

        stdout = self.comm_get_snapshot(ldev)
        if stdout:
            lines = stdout.splitlines()
            line = shlex.split(lines[1])
            status = STATUS_TABLE.get(line[2], basic_lib.UNKN)

            if line[1] == 'P-VOL':
                pvol = ldev
                svol = int(line[6])
            else:
                is_svol = True
                pvol = int(line[6])
                svol = ldev

                if status == basic_lib.PSUS:
                    status = self.get_snap_pvol_status(pvol, svol)

            svol_info = {'lun': svol, 'status': status, 'is_vvol': True}
            paired_info['svol'].append(svol_info)
            paired_info['pvol'] = pvol

        if only_flag or is_svol:
            return paired_info

        for copy_group in self.get_copy_groups():
            ldev_name = None
            pvol_status = basic_lib.UNKN
            svol_status = basic_lib.UNKN

            ret, stdout, stderr = self.comm_pairdisplay(copy_group)
            if not stdout:
                continue

            lines = stdout.splitlines()
            for line in lines[1:]:
                line = shlex.split(line)
                if line[9] not in ['P-VOL', 'S-VOL']:
                    continue

                ldev0 = int(line[8])
                ldev1 = int(line[12])
                if ldev not in [ldev0, ldev1]:
                    continue

                ldev_name = line[1]

                if line[9] == 'P-VOL':
                    pvol = ldev0
                    svol = ldev1
                    pvol_status = STATUS_TABLE.get(line[10], basic_lib.UNKN)
                else:
                    svol = ldev0
                    pvol = ldev1
                    svol_status = STATUS_TABLE.get(line[10], basic_lib.UNKN)

                if svol == ldev:
                    is_svol = True

            if not ldev_name:
                continue

            pvol_group = '%sP' % copy_group
            pvol_ok = self.check_device_grp(pvol_group, pvol,
                                            ldev_name=ldev_name)

            svol_group = '%sS' % copy_group
            svol_ok = self.check_device_grp(svol_group, svol,
                                            ldev_name=ldev_name)

            if pvol_ok and svol_ok:
                if pvol_status == basic_lib.PSUS:
                    status = svol_status
                else:
                    status = pvol_status

                svol_info = {'lun': svol, 'status': status, 'is_vvol': False}
                paired_info['svol'].append(svol_info)

                if is_svol:
                    break

        # When 'pvol' is 0, it should be true.
        # So, it cannot remove 'is not None'.
        if pvol is not None and paired_info['pvol'] is None:
            paired_info['pvol'] = pvol

        return paired_info

    def add_pair_config(self, pvol, svol, copy_group, ldev_name, mun):
        pvol_group = '%sP' % copy_group
        svol_group = '%sS' % copy_group
        self.comm_add_device_grp(pvol_group, ldev_name, pvol)
        self.comm_add_device_grp(svol_group, ldev_name, svol)
        nr_copy_groups = self.check_copy_grp(copy_group)
        if nr_copy_groups == 1:
            self.comm_delete_copy_grp(copy_group)
        if nr_copy_groups != 2:
            self.comm_add_copy_grp(copy_group, pvol_group, svol_group, mun)

    def delete_pair_config(self, pvol, svol, copy_group, ldev_name):
        pvol_group = '%sP' % copy_group
        svol_group = '%sS' % copy_group
        if self.check_device_grp(pvol_group, pvol, ldev_name=ldev_name):
            self.comm_delete_device_grp(pvol_group, pvol)
        if self.check_device_grp(svol_group, svol, ldev_name=ldev_name):
            self.comm_delete_device_grp(svol_group, svol)

    def _wait_for_pair_status(self, copy_group, ldev_name,
                              status, timeout, check_svol, start):
        if self.comm_pairevtwait(copy_group, ldev_name,
                                 check_svol) in status:
            raise loopingcall.LoopingCallDone()

        if time.time() - start >= timeout:
            msg = basic_lib.output_err(
                637, method='_wait_for_pair_status', timout=timeout)
            raise exception.HBSDError(message=msg)

    def wait_pair(self, copy_group, ldev_name, status, timeout,
                  interval, check_svol=False):
        loop = loopingcall.FixedIntervalLoopingCall(
            self._wait_for_pair_status, copy_group, ldev_name,
            status, timeout, check_svol, time.time())

        loop.start(interval=interval).wait()

    def comm_create_pair(self, pvol, svol, is_vvol):
        timeout = basic_lib.DEFAULT_PROCESS_WAITTIME
        interval = self.conf.hitachi_copy_check_interval
        if not is_vvol:
            restart = False
            create = False
            ldev_name = LDEV_NAME % (pvol, svol)
            mun = 0
            for mun in range(MAX_MUNS):
                copy_group = self.copy_groups[mun]
                pvol_group = '%sP' % copy_group

                if not self.check_device_grp(pvol_group, pvol):
                    break
            else:
                msg = basic_lib.output_err(
                    615, copy_method=basic_lib.FULL, pvol=pvol)
                raise exception.HBSDBusy(message=msg)
            try:
                self.add_pair_config(pvol, svol, copy_group, ldev_name, mun)
                self.restart_pair_horcm()
                restart = True
                self.comm_paircreate(copy_group, ldev_name)
                create = True
                self.wait_pair(copy_group, ldev_name, [basic_lib.PSUS],
                               timeout, interval)
                self.wait_pair(copy_group, ldev_name,
                               [basic_lib.PSUS, basic_lib.COPY],
                               timeout, interval, check_svol=True)
            except Exception:
                with excutils.save_and_reraise_exception():
                    if create:
                        try:
                            self.wait_pair(copy_group, ldev_name,
                                           [basic_lib.PSUS], timeout,
                                           interval)
                            self.wait_pair(copy_group, ldev_name,
                                           [basic_lib.PSUS], timeout,
                                           interval, check_svol=True)
                        except Exception as ex:
                            LOG.warning(_LW('Failed to create pair: %s'), ex)

                        try:
                            self.comm_pairsplit(copy_group, ldev_name)
                            self.wait_pair(
                                copy_group, ldev_name,
                                [basic_lib.SMPL], timeout,
                                self.conf.hitachi_async_copy_check_interval)
                        except Exception as ex:
                            LOG.warning(_LW('Failed to create pair: %s'), ex)

                    if self.is_smpl(copy_group, ldev_name):
                        try:
                            self.delete_pair_config(pvol, svol, copy_group,
                                                    ldev_name)
                        except Exception as ex:
                            LOG.warning(_LW('Failed to create pair: %s'), ex)

                    if restart:
                        try:
                            self.restart_pair_horcm()
                        except Exception as ex:
                            LOG.warning(_LW('Failed to restart horcm: %s'), ex)

        else:
            self.check_snap_count(pvol)
            self.comm_add_snapshot(pvol, svol)

            try:
                self.wait_snap(pvol, svol, [basic_lib.PAIR], timeout, interval)
                self.comm_modify_snapshot(svol, 'create')
                self.wait_snap(pvol, svol, [basic_lib.PSUS], timeout, interval)
            except Exception:
                with excutils.save_and_reraise_exception():
                    try:
                        self.comm_delete_snapshot(svol)
                        self.wait_snap(
                            pvol, svol, [basic_lib.SMPL], timeout,
                            self.conf.hitachi_async_copy_check_interval)
                    except Exception as ex:
                        LOG.warning(_LW('Failed to create pair: %s'), ex)

    def delete_pair(self, pvol, svol, is_vvol):
        timeout = basic_lib.DEFAULT_PROCESS_WAITTIME
        interval = self.conf.hitachi_async_copy_check_interval
        if not is_vvol:
            ldev_name = LDEV_NAME % (pvol, svol)
            copy_group = self.get_matched_copy_group(pvol, svol, ldev_name)
            if not copy_group:
                return
            try:
                self.comm_pairsplit(copy_group, ldev_name)
                self.wait_pair(copy_group, ldev_name, [basic_lib.SMPL],
                               timeout, interval)
            finally:
                if self.is_smpl(copy_group, ldev_name):
                    self.delete_pair_config(pvol, svol, copy_group, ldev_name)
        else:
            self.comm_delete_snapshot(svol)
            self.wait_snap(pvol, svol, [basic_lib.SMPL], timeout, interval)

    def comm_raidqry(self):
        ret, stdout, stderr = self.exec_command('raidqry', '-h')
        if ret:
            opt = 'raidqry -h'
            msg = basic_lib.output_err(
                600, cmd=opt, ret=ret, out=stdout, err=stderr)
            raise exception.HBSDCmdError(message=msg, ret=ret, err=stderr)
        return stdout

    def get_comm_version(self):
        stdout = self.comm_raidqry()
        lines = stdout.splitlines()
        return shlex.split(lines[1])[1]

    def output_param_to_log(self, conf):
        for opt in volume_opts:
            if not opt.secret:
                value = getattr(conf, opt.name)
                LOG.info(_LI('\t%(name)-35s : %(value)s'),
                         {'name': opt.name, 'value': value})

    def create_lock_file(self):
        inst = self.conf.hitachi_horcm_numbers[0]
        pair_inst = self.conf.hitachi_horcm_numbers[1]
        serial = self.conf.hitachi_serial_number
        raidcom_lock_file = '%s%d' % (RAIDCOM_LOCK_FILE, inst)
        raidcom_pair_lock_file = '%s%d' % (RAIDCOM_LOCK_FILE, pair_inst)
        horcmgr_lock_file = '%s%d' % (HORCMGR_LOCK_FILE, pair_inst)
        resource_lock_file = '%s%s' % (RESOURCE_LOCK_FILE, serial)

        basic_lib.create_empty_file(raidcom_lock_file)
        basic_lib.create_empty_file(raidcom_pair_lock_file)
        basic_lib.create_empty_file(horcmgr_lock_file)
        basic_lib.create_empty_file(resource_lock_file)

    def connect_storage(self):
        properties = utils.brick_get_connector_properties()
        self.setup_horcmgr(properties['ip'])

    def get_max_hostgroups(self):
        """return the maximum value of hostgroup id."""
        return MAX_HOSTGROUPS

    def get_hostgroup_luns(self, port, gid):
        list = []
        self.add_used_hlun(port, gid, list)

        return list

    def get_ldev_size_in_gigabyte(self, ldev, existing_ref):
        param = 'serial_number'

        if param not in existing_ref:
            msg = basic_lib.output_err(700, param=param)
            raise exception.HBSDError(data=msg)

        storage = existing_ref.get(param)
        if storage != self.conf.hitachi_serial_number:
            msg = basic_lib.output_err(648, resource=param)
            raise exception.HBSDError(data=msg)

        stdout = self.comm_get_ldev(ldev)
        if not stdout:
            msg = basic_lib.output_err(648, resource='LDEV')
            raise exception.HBSDError(data=msg)

        sts_line = vol_type = ""
        vol_attrs = []
        size = num_port = 1

        lines = stdout.splitlines()
        for line in lines:
            if line.startswith("STS :"):
                sts_line = line

            elif line.startswith("VOL_TYPE :"):
                vol_type = shlex.split(line)[2]

            elif line.startswith("VOL_ATTR :"):
                vol_attrs = shlex.split(line)[2:]

            elif line.startswith("VOL_Capacity(BLK) :"):
                size = int(shlex.split(line)[2])

            elif line.startswith("NUM_PORT :"):
                num_port = int(shlex.split(line)[2])

        if 'NML' not in sts_line:
            msg = basic_lib.output_err(648, resource='LDEV')

            raise exception.HBSDError(data=msg)

        if 'OPEN-V' not in vol_type:
            msg = basic_lib.output_err(702, ldev=ldev)
            raise exception.HBSDError(data=msg)

        if 'HDP' not in vol_attrs:
            msg = basic_lib.output_err(702, ldev=ldev)
            raise exception.HBSDError(data=msg)

        for vol_attr in vol_attrs:
            if vol_attr == ':':
                continue

            if vol_attr in PAIR_TYPE:
                msg = basic_lib.output_err(705, ldev=ldev)
                raise exception.HBSDError(data=msg)

            if vol_attr not in PERMITTED_TYPE:
                msg = basic_lib.output_err(702, ldev=ldev)
                raise exception.HBSDError(data=msg)

        # Hitachi storage calculates volume sizes in a block unit, 512 bytes.
        # So, units.Gi is divided by 512.
        if size % (units.Gi / 512):
            msg = basic_lib.output_err(703, ldev=ldev)
            raise exception.HBSDError(data=msg)

        if num_port:
            msg = basic_lib.output_err(704, ldev=ldev)
            raise exception.HBSDError(data=msg)

        return size / (units.Gi / 512)



# Copyright (c) 2014 NetApp, Inc
# Copyright (c) 2014 Navneet Singh
# Copyright (c) 2015 Alex Meade
# Copyright (c) 2015 Rushil Chugh
# Copyright (c) 2015 Yogesh Kshirsagar
# Copyright (c) 2015 Michael Price
#  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Client classes for web services.
"""

import copy
import json
import uuid

from oslo_log import log as logging
import requests
from simplejson import scanner
import six
from six.moves import urllib

from cinder import exception
from cinder.i18n import _
from cinder.i18n import _LE
import cinder.utils as cinder_utils
from cinder.volume.drivers.netapp.eseries import exception as es_exception
from cinder.volume.drivers.netapp.eseries import utils
from cinder.volume.drivers.netapp import utils as na_utils


LOG = logging.getLogger(__name__)


class WebserviceClient(object):
    """Base client for NetApp Storage web services."""

    def __init__(self, scheme, host, port, service_path, username,
                 password, **kwargs):
        self._validate_params(scheme, host, port)
        self._create_endpoint(scheme, host, port, service_path)
        self._username = username
        self._password = password
        self._init_connection()

    def _validate_params(self, scheme, host, port):
        """Does some basic validation for web service params."""
        if host is None or port is None or scheme is None:
            msg = _('One of the required inputs from host, '
                    'port or scheme was not found.')
            raise exception.InvalidInput(reason=msg)
        if scheme not in ('http', 'https'):
            raise exception.InvalidInput(reason=_("Invalid transport type."))

    def _create_endpoint(self, scheme, host, port, service_path):
        """Creates end point url for the service."""
        netloc = '%s:%s' % (host, port)
        self._endpoint = urllib.parse.urlunparse((scheme, netloc, service_path,
                                                 None, None, None))

    def _init_connection(self):
        """Do client specific set up for session and connection pooling."""
        self.conn = requests.Session()
        if self._username and self._password:
            self.conn.auth = (self._username, self._password)

    def invoke_service(self, method='GET', url=None, params=None, data=None,
                       headers=None, timeout=None, verify=False):
        url = url or self._endpoint
        try:
            response = self.conn.request(method, url, params, data,
                                         headers=headers, timeout=timeout,
                                         verify=verify)
        # Catching error conditions other than the perceived ones.
        # Helps propagating only known exceptions back to the caller.
        except Exception as e:
            LOG.exception(_LE("Unexpected error while invoking web service."
                              " Error - %s."), e)
            raise exception.NetAppDriverException(
                _("Invoking web service failed."))
        return response


class RestClient(WebserviceClient):
    """REST client specific to e-series storage service."""

    ID = 'id'
    WWN = 'worldWideName'
    NAME = 'label'

    ASUP_VALID_VERSION = (1, 52, 9000, 3)
    # We need to check for both the release and the pre-release versions
    SSC_VALID_VERSIONS = ((1, 53, 9000, 1), (1, 53, 9010, 17))
    REST_1_3_VERSION = (1, 53, 9000, 1)
    REST_1_4_VERSIONS = ((1, 54, 9000, 1), (1, 54, 9090, 0))

    RESOURCE_PATHS = {
        'volumes': '/storage-systems/{system-id}/volumes',
        'volume': '/storage-systems/{system-id}/volumes/{object-id}',
        'pool_operation_progress':
            '/storage-systems/{system-id}/storage-pools/{object-id}'
            '/action-progress',
        'volume_expand':
            '/storage-systems/{system-id}/volumes/{object-id}/expand',
        'thin_volume_expand':
            '/storage-systems/{system-id}/thin-volumes/{object-id}/expand',
        'ssc_volumes': '/storage-systems/{system-id}/ssc/volumes',
        'ssc_volume': '/storage-systems/{system-id}/ssc/volumes/{object-id}',
        'snapshot_groups': '/storage-systems/{system-id}/snapshot-groups',
        'snapshot_group':
            '/storage-systems/{system-id}/snapshot-groups/{object-id}',
        'snapshot_volumes': '/storage-systems/{system-id}/snapshot-volumes',
        'snapshot_volume':
            '/storage-systems/{system-id}/snapshot-volumes/{object-id}',
        'snapshot_images': '/storage-systems/{system-id}/snapshot-images',
        'snapshot_image':
            '/storage-systems/{system-id}/snapshot-images/{object-id}',
        'cgroup':
            '/storage-systems/{system-id}/consistency-groups/{object-id}',
        'cgroups': '/storage-systems/{system-id}/consistency-groups',
        'cgroup_members':
            '/storage-systems/{system-id}/consistency-groups/{object-id}'
            '/member-volumes',
        'cgroup_member':
            '/storage-systems/{system-id}/consistency-groups/{object-id}'
            '/member-volumes/{vol-id}',
        'cgroup_snapshots':
            '/storage-systems/{system-id}/consistency-groups/{object-id}'
            '/snapshots',
        'cgroup_snapshot':
            '/storage-systems/{system-id}/consistency-groups/{object-id}'
            '/snapshots/{seq-num}',
        'cgroup_snapshots_by_seq':
            '/storage-systems/{system-id}/consistency-groups/{object-id}'
            '/snapshots/{seq-num}',
        'cgroup_cgsnap_view':
            '/storage-systems/{system-id}/consistency-groups/{object-id}'
            '/views/{seq-num}',
        'cgroup_cgsnap_views':
            '/storage-systems/{system-id}/consistency-groups/{object-id}'
            '/views/',
        'cgroup_snapshot_views':
            '/storage-systems/{system-id}/consistency-groups/{object-id}'
            '/views/{view-id}/views',
        'persistent-stores': '/storage-systems/{'
                             'system-id}/persistent-records/',
        'persistent-store': '/storage-systems/{'
                            'system-id}/persistent-records/{key}'
    }

    def __init__(self, scheme, host, port, service_path, username,
                 password, **kwargs):

        super(RestClient, self).__init__(scheme, host, port, service_path,
                                         username, password, **kwargs)

        kwargs = kwargs or {}

        self._system_id = kwargs.get('system_id')
        self._content_type = kwargs.get('content_type') or 'json'

    def _init_features(self):
        """Sets up and initializes E-Series feature support map."""
        self.features = na_utils.Features()
        self.api_operating_mode, self.api_version = self.get_eseries_api_info(
            verify=False)

        api_version_tuple = tuple(int(version)
                                  for version in self.api_version.split('.'))

        asup_api_valid_version = self._validate_version(
            self.ASUP_VALID_VERSION, api_version_tuple)

        rest_1_3_api_valid_version = self._validate_version(
            self.REST_1_3_VERSION, api_version_tuple)

        rest_1_4_api_valid_version = any(
            self._validate_version(valid_version, api_version_tuple)
            for valid_version in self.REST_1_4_VERSIONS)

        ssc_api_valid_version = any(self._validate_version(valid_version,
                                                           api_version_tuple)
                                    for valid_version
                                    in self.SSC_VALID_VERSIONS)

        self.features.add_feature('AUTOSUPPORT',
                                  supported=asup_api_valid_version,
                                  min_version=self._version_tuple_to_str(
                                      self.ASUP_VALID_VERSION))
        self.features.add_feature('SSC_API_V2',
                                  supported=ssc_api_valid_version,
                                  min_version=self._version_tuple_to_str(
                                      self.SSC_VALID_VERSIONS[0]))
        self.features.add_feature(
            'REST_1_3_RELEASE', supported=rest_1_3_api_valid_version,
            min_version=self._version_tuple_to_str(self.REST_1_3_VERSION))
        self.features.add_feature(
            'REST_1_4_RELEASE', supported=rest_1_4_api_valid_version,
            min_version=self._version_tuple_to_str(self.REST_1_4_VERSIONS[0]))

    def _version_tuple_to_str(self, version):
        return ".".join([str(part) for part in version])

    def _validate_version(self, version, actual_version):
        """Determine if version is newer than, or equal to the actual version

        The proxy version number is formatted as AA.BB.CCCC.DDDD
        A: Major version part 1
        B: Major version part 2
        C: Release version: 9000->Release, 9010->Pre-release, 9090->Integration
        D: Minor version

        Examples:
        02.53.9000.0010
        02.52.9010.0001

        Note: The build version is actually 'newer' the lower the release
        (CCCC) number is.

        :param version: The version to validate
        :param actual_version: The running version of the Webservice
        :returns: True if the actual_version is equal or newer than the
        current running version, otherwise False
        """
        major_1, major_2, release, minor = version
        actual_major_1, actual_major_2, actual_release, actual_minor = (
            actual_version)

        # We need to invert the release number for it to work with this
        # comparison
        return (actual_major_1, actual_major_2, 10000 - actual_release,
                actual_minor) >= (major_1, major_2, 10000 - release, minor)

    def set_system_id(self, system_id):
        """Set the storage system id."""
        self._system_id = system_id

    def get_system_id(self):
        """Get the storage system id."""
        return getattr(self, '_system_id', None)

    def _get_resource_url(self, path, use_system=True, **kwargs):
        """Creates end point url for rest service."""
        kwargs = kwargs or {}
        if use_system:
            if not self._system_id:
                raise exception.NotFound(_('Storage system id not set.'))
            kwargs['system-id'] = self._system_id
        path = path.format(**kwargs)
        if not self._endpoint.endswith('/'):
            self._endpoint = '%s/' % self._endpoint
        return urllib.parse.urljoin(self._endpoint, path.lstrip('/'))

    def _invoke(self, method, path, data=None, use_system=True,
                timeout=None, verify=False, **kwargs):
        """Invokes end point for resource on path."""
        url = self._get_resource_url(path, use_system, **kwargs)
        if self._content_type == 'json':
            headers = {'Accept': 'application/json',
                       'Content-Type': 'application/json'}
            if cinder_utils.TRACE_API:
                self._log_http_request(method, url, headers, data)
            data = json.dumps(data) if data else None
            res = self.invoke_service(method, url, data=data,
                                      headers=headers,
                                      timeout=timeout, verify=verify)

            try:
                res_dict = res.json() if res.text else None
            # This should only occur if we expected JSON, but were sent
            # something else
            except scanner.JSONDecodeError:
                res_dict = None

            if cinder_utils.TRACE_API:
                self._log_http_response(res.status_code, dict(res.headers),
                                        res_dict)

            self._eval_response(res)
            return res_dict
        else:
            raise exception.NetAppDriverException(
                _("Content type not supported."))

    def _to_pretty_dict_string(self, data):
        """Convert specified dict to pretty printed string."""
        return json.dumps(data, sort_keys=True,
                          indent=2, separators=(',', ': '))

    def _log_http_request(self, verb, url, headers, body):
        scrubbed_body = copy.deepcopy(body)
        if scrubbed_body:
            if 'password' in scrubbed_body:
                scrubbed_body['password'] = "****"
            if 'storedPassword' in scrubbed_body:
                scrubbed_body['storedPassword'] = "****"

        params = {'verb': verb, 'path': url,
                  'body': self._to_pretty_dict_string(scrubbed_body) or "",
                  'headers': self._to_pretty_dict_string(headers)}
        LOG.debug("Invoking ESeries Rest API, Request:\n"
                  "HTTP Verb: %(verb)s\n"
                  "URL Path: %(path)s\n"
                  "HTTP Headers:\n"
                  "%(headers)s\n"
                  "Body:\n"
                  "%(body)s\n", (params))

    def _log_http_response(self, status, headers, body):
        params = {'status': status,
                  'body': self._to_pretty_dict_string(body) or "",
                  'headers': self._to_pretty_dict_string(headers)}
        LOG.debug("ESeries Rest API, Response:\n"
                  "HTTP Status Code: %(status)s\n"
                  "HTTP Headers:\n"
                  "%(headers)s\n"
                  "Body:\n"
                  "%(body)s\n", (params))

    def _eval_response(self, response):
        """Evaluates response before passing result to invoker."""
        status_code = int(response.status_code)
        # codes >= 300 are not ok and to be treated as errors
        if status_code >= 300:
            # Response code 422 returns error code and message
            if status_code == 422:
                msg = _("Response error - %s.") % response.text
                json_response = response.json()
                if json_response is not None:
                    ret_code = json_response.get('retcode', '')
                    if ret_code == '30' or ret_code == 'authFailPassword':
                        msg = _("The storage array password for %s is "
                                "incorrect, please update the configured "
                                "password.") % self._system_id
            elif status_code == 424:
                msg = _("Response error - The storage-system is offline.")
            else:
                msg = _("Response error code - %s.") % status_code
            raise es_exception.WebServiceException(msg,
                                                   status_code=status_code)

    def _get_volume_api_path(self, path_key):
        """Retrieve the correct API path based on API availability

        :param path_key: The volume API to request (volume or volumes)
        :raise KeyError: If the path_key is not valid
        """
        if self.features.SSC_API_V2:
            path_key = 'ssc_' + path_key
        return self.RESOURCE_PATHS[path_key]

    def create_volume(self, pool, label, size, unit='gb', seg_size=0,
                      read_cache=None, write_cache=None, flash_cache=None,
                      data_assurance=None, thin_provision=False):
        """Creates a volume on array with the configured attributes

        Note: if read_cache, write_cache, flash_cache, or data_assurance
         are not provided, the default will be utilized by the Webservice.

        :param pool: The pool unique identifier
        :param label: The unqiue label for the volume
        :param size: The capacity in units
        :param unit: The unit for capacity
        :param seg_size: The segment size for the volume, expressed in KB.
        Default will allow the Webservice to choose.
        :param read_cache: If true, enable read caching, if false,
        explicitly disable it.
        :param write_cache: If true, enable write caching, if false,
        explicitly disable it.
        :param flash_cache: If true, add the volume to a Flash Cache
        :param data_assurance: If true, enable the Data Assurance capability
        :returns: The created volume
        """

        # Utilize the new API if it is available
        if self.features.SSC_API_V2:
            path = "/storage-systems/{system-id}/ssc/volumes"
            data = {'poolId': pool, 'name': label, 'sizeUnit': unit,
                    'size': int(size), 'dataAssuranceEnable': data_assurance,
                    'flashCacheEnable': flash_cache,
                    'readCacheEnable': read_cache,
                    'writeCacheEnable': write_cache,
                    'thinProvision': thin_provision}
        # Use the old API
        else:
            # Determine if there are were extra specs provided that are not
            # supported
            extra_specs = [read_cache, write_cache]
            unsupported_spec = any([spec is not None for spec in extra_specs])
            if(unsupported_spec):
                msg = _("E-series proxy API version %(current_version)s does "
                        "not support full set of SSC extra specs. The proxy"
                        " version must be at at least %(min_version)s.")
                min_version = self.features.SSC_API_V2.minimum_version
                raise exception.NetAppDriverException(msg %
                                                      {'current_version':
                                                       self.api_version,
                                                       'min_version':
                                                       min_version})

            path = "/storage-systems/{system-id}/volumes"
            data = {'poolId': pool, 'name': label, 'sizeUnit': unit,
                    'size': int(size), 'segSize': seg_size}
        return self._invoke('POST', path, data)

    def delete_volume(self, object_id):
        """Deletes given volume from array."""
        if self.features.SSC_API_V2:
            path = self.RESOURCE_PATHS.get('ssc_volume')
        else:
            path = self.RESOURCE_PATHS.get('volume')
        return self._invoke('DELETE', path, **{'object-id': object_id})

    def list_volumes(self):
        """Lists all volumes in storage array."""
        if self.features.SSC_API_V2:
            path = self.RESOURCE_PATHS.get('ssc_volumes')
        else:
            path = self.RESOURCE_PATHS.get('volumes')

        return self._invoke('GET', path)

    def list_volume(self, object_id):
        """Retrieve the given volume from array.

        :param object_id: The volume id, label, or wwn
        :returns: The volume identified by object_id
        :raise: VolumeNotFound if the volume could not be found
        """

        if self.features.SSC_API_V2:
            return self._list_volume_v2(object_id)
        # The new API is not available,
        else:
            # Search for the volume with label, id, or wwn.
            return self._list_volume_v1(object_id)

    def _list_volume_v1(self, object_id):
        # Search for the volume with label, id, or wwn.
        for vol in self.list_volumes():
            if (object_id == vol.get(self.NAME) or object_id == vol.get(
                    self.WWN) or object_id == vol.get(self.ID)):
                return vol
        # The volume could not be found
        raise exception.VolumeNotFound(volume_id=object_id)

    def _list_volume_v2(self, object_id):
        path = self.RESOURCE_PATHS.get('ssc_volume')
        try:
            return self._invoke('GET', path, **{'object-id': object_id})
        except es_exception.WebServiceException as e:
            if 404 == e.status_code:
                raise exception.VolumeNotFound(volume_id=object_id)
            else:
                raise

    def update_volume(self, object_id, label):
        """Renames given volume on array."""
        if self.features.SSC_API_V2:
            path = self.RESOURCE_PATHS.get('ssc_volume')
        else:
            path = self.RESOURCE_PATHS.get('volume')
        data = {'name': label}
        return self._invoke('POST', path, data, **{'object-id': object_id})

    def create_consistency_group(self, name, warn_at_percent_full=75,
                                 rollback_priority='medium',
                                 full_policy='failbasewrites'):
        """Define a new consistency group"""
        path = self.RESOURCE_PATHS.get('cgroups')
        data = {
            'name': name,
            'fullWarnThresholdPercent': warn_at_percent_full,
            'repositoryFullPolicy': full_policy,
            # A non-zero threshold enables auto-deletion
            'autoDeleteThreshold': 0,
            'rollbackPriority': rollback_priority,
        }

        return self._invoke('POST', path, data)

    def get_consistency_group(self, object_id):
        """Retrieve the consistency group identified by object_id"""
        path = self.RESOURCE_PATHS.get('cgroup')

        return self._invoke('GET', path, **{'object-id': object_id})

    def list_consistency_groups(self):
        """Retrieve all consistency groups defined on the array"""
        path = self.RESOURCE_PATHS.get('cgroups')

        return self._invoke('GET', path)

    def delete_consistency_group(self, object_id):
        path = self.RESOURCE_PATHS.get('cgroup')

        self._invoke('DELETE', path, **{'object-id': object_id})

    def add_consistency_group_member(self, volume_id, cg_id,
                                     repo_percent=20.0):
        """Add a volume to a consistency group

        :param volume_id the eseries volume id
        :param cg_id: the eseries cg id
        :param repo_percent: percentage capacity of the volume to use for
        capacity of the copy-on-write repository
        """
        path = self.RESOURCE_PATHS.get('cgroup_members')
        data = {'volumeId': volume_id, 'repositoryPercent': repo_percent}

        return self._invoke('POST', path, data, **{'object-id': cg_id})

    def remove_consistency_group_member(self, volume_id, cg_id):
        """Remove a volume from a consistency group"""
        path = self.RESOURCE_PATHS.get('cgroup_member')

        self._invoke('DELETE', path, **{'object-id': cg_id,
                                        'vol-id': volume_id})

    def create_consistency_group_snapshot(self, cg_id):
        """Define a consistency group snapshot"""
        path = self.RESOURCE_PATHS.get('cgroup_snapshots')

        return self._invoke('POST', path, **{'object-id': cg_id})

    def delete_consistency_group_snapshot(self, cg_id, seq_num):
        """Define a consistency group snapshot"""
        path = self.RESOURCE_PATHS.get('cgroup_snapshot')

        return self._invoke('DELETE', path, **{'object-id': cg_id,
                                               'seq-num': seq_num})

    def get_consistency_group_snapshots(self, cg_id):
        """Retrieve all snapshots defined for a consistency group"""
        path = self.RESOURCE_PATHS.get('cgroup_snapshots')

        return self._invoke('GET', path, **{'object-id': cg_id})

    def create_cg_snapshot_view(self, cg_id, name, snap_id):
        """Define a snapshot view for the cgsnapshot

        In order to define a snapshot view for a snapshot defined under a
        consistency group, the view must be defined at the cgsnapshot
        level.

        :param cg_id: E-Series cg identifier
        :param name: the label for the view
        :param snap_id: E-Series snapshot view to locate
        :raise NetAppDriverException: if the snapshot view cannot be
        located for the snapshot identified by snap_id
        :return snapshot view for snapshot identified by snap_id
        """
        path = self.RESOURCE_PATHS.get('cgroup_cgsnap_views')

        data = {
            'name': name,
            'accessMode': 'readOnly',
            # Only define a view for this snapshot
            'pitId': snap_id,
        }
        # Define a view for the cgsnapshot
        cgsnapshot_view = self._invoke(
            'POST', path, data, **{'object-id': cg_id})

        # Retrieve the snapshot views associated with our cgsnapshot view
        views = self.list_cg_snapshot_views(cg_id, cgsnapshot_view[
            'cgViewRef'])
        # Find the snapshot view defined for our snapshot
        for view in views:
            if view['basePIT'] == snap_id:
                return view
        else:
            try:
                self.delete_cg_snapshot_view(cg_id, cgsnapshot_view['id'])
            finally:
                raise exception.NetAppDriverException(
                    'Unable to create snapshot view.')

    def list_cg_snapshot_views(self, cg_id, view_id):
        path = self.RESOURCE_PATHS.get('cgroup_snapshot_views')

        return self._invoke('GET', path, **{'object-id': cg_id,
                                            'view-id': view_id})

    def delete_cg_snapshot_view(self, cg_id, view_id):
        path = self.RESOURCE_PATHS.get('cgroup_snap_view')

        return self._invoke('DELETE', path, **{'object-id': cg_id,
                                               'view-id': view_id})

    def get_pool_operation_progress(self, object_id):
        """Retrieve the progress long-running operations on a storage pool

        Example:
        [
            {
                "volumeRef": "3232....", # Volume being referenced
                "progressPercentage": 0, # Approxmate percent complete
                "estimatedTimeToCompletion": 0, # ETA in minutes
                "currentAction": "none" # Current volume action
            }
            ...
        ]

        :param object_id: A pool id
        :returns: A dict representing the action progress
        """
        path = self.RESOURCE_PATHS.get('pool_operation_progress')
        return self._invoke('GET', path, **{'object-id': object_id})

    def expand_volume(self, object_id, new_capacity, thin_provisioned,
                      capacity_unit='gb'):
        """Increase the capacity of a volume"""
        if thin_provisioned:
            path = self.RESOURCE_PATHS.get('thin_volume_expand')
            data = {'newVirtualSize': new_capacity, 'sizeUnit': capacity_unit,
                    'newRepositorySize': new_capacity}
            return self._invoke('POST', path, data, **{'object-id': object_id})
        else:
            path = self.RESOURCE_PATHS.get('volume_expand')
            data = {'expansionSize': new_capacity, 'sizeUnit': capacity_unit}
            return self._invoke('POST', path, data, **{'object-id': object_id})

    def get_volume_mappings(self):
        """Creates volume mapping on array."""
        path = "/storage-systems/{system-id}/volume-mappings"
        return self._invoke('GET', path)

    def get_volume_mappings_for_volume(self, volume):
        """Gets all host mappings for given volume from array."""
        mappings = self.get_volume_mappings() or []
        return [x for x in mappings
                if x.get('volumeRef') == volume['volumeRef']]

    def get_volume_mappings_for_host(self, host_ref):
        """Gets all volume mappings for given host from array."""
        mappings = self.get_volume_mappings() or []
        return [x for x in mappings if x.get('mapRef') == host_ref]

    def get_volume_mappings_for_host_group(self, hg_ref):
        """Gets all volume mappings for given host group from array."""
        mappings = self.get_volume_mappings() or []
        return [x for x in mappings if x.get('mapRef') == hg_ref]

    def create_volume_mapping(self, object_id, target_id, lun):
        """Creates volume mapping on array."""
        path = "/storage-systems/{system-id}/volume-mappings"
        data = {'mappableObjectId': object_id, 'targetId': target_id,
                'lun': lun}
        return self._invoke('POST', path, data)

    def delete_volume_mapping(self, map_object_id):
        """Deletes given volume mapping from array."""
        path = "/storage-systems/{system-id}/volume-mappings/{object-id}"
        return self._invoke('DELETE', path, **{'object-id': map_object_id})

    def move_volume_mapping_via_symbol(self, map_ref, to_ref, lun_id):
        """Moves a map from one host/host_group object to another."""

        path = "/storage-systems/{system-id}/symbol/moveLUNMapping"
        data = {'lunMappingRef': map_ref,
                'lun': int(lun_id),
                'mapRef': to_ref}
        return_code = self._invoke('POST', path, data)
        if return_code == 'ok':
            return {'lun': lun_id}
        msg = _("Failed to move LUN mapping.  Return code: %s") % return_code
        raise exception.NetAppDriverException(msg)

    def list_hardware_inventory(self):
        """Lists objects in the hardware inventory."""
        path = "/storage-systems/{system-id}/hardware-inventory"
        return self._invoke('GET', path)

    def list_target_wwpns(self):
        """Lists the world-wide port names of the target."""
        inventory = self.list_hardware_inventory()
        fc_ports = inventory.get("fibrePorts", [])
        wwpns = [port['portName'] for port in fc_ports]
        return wwpns

    def create_host_group(self, label):
        """Creates a host group on the array."""
        path = "/storage-systems/{system-id}/host-groups"
        data = {'name': label}
        return self._invoke('POST', path, data)

    def get_host_group(self, host_group_ref):
        """Gets a single host group from the array."""
        path = "/storage-systems/{system-id}/host-groups/{object-id}"
        try:
            return self._invoke('GET', path, **{'object-id': host_group_ref})
        except exception.NetAppDriverException:
            raise exception.NotFound(_("Host group with ref %s not found") %
                                     host_group_ref)

    def get_host_group_by_name(self, name):
        """Gets a single host group by name from the array."""
        host_groups = self.list_host_groups()
        matching = [host_group for host_group in host_groups
                    if host_group['label'] == name]
        if len(matching):
            return matching[0]
        raise exception.NotFound(_("Host group with name %s not found") % name)

    def list_host_groups(self):
        """Lists host groups on the array."""
        path = "/storage-systems/{system-id}/host-groups"
        return self._invoke('GET', path)

    def list_hosts(self):
        """Lists host objects in the system."""
        path = "/storage-systems/{system-id}/hosts"
        return self._invoke('GET', path)

    def create_host(self, label, host_type, ports=None, group_id=None):
        """Creates host on array."""
        path = "/storage-systems/{system-id}/hosts"
        data = {'name': label, 'hostType': host_type}
        data.setdefault('groupId', group_id if group_id else None)
        data.setdefault('ports', ports if ports else None)
        return self._invoke('POST', path, data)

    def create_host_with_ports(self, label, host_type, port_ids,
                               port_type='iscsi', group_id=None):
        """Creates host on array with given port information."""
        if port_type == 'fc':
            port_ids = [six.text_type(wwpn).replace(':', '')
                        for wwpn in port_ids]
        ports = []
        for port_id in port_ids:
            port_label = utils.convert_uuid_to_es_fmt(uuid.uuid4())
            port = {'type': port_type, 'port': port_id, 'label': port_label}
            ports.append(port)
        return self.create_host(label, host_type, ports, group_id)

    def update_host(self, host_ref, data):
        """Updates host type for a given host."""
        path = "/storage-systems/{system-id}/hosts/{object-id}"
        return self._invoke('POST', path, data, **{'object-id': host_ref})

    def get_host(self, host_ref):
        """Gets a single host from the array."""
        path = "/storage-systems/{system-id}/hosts/{object-id}"
        return self._invoke('GET', path, **{'object-id': host_ref})

    def update_host_type(self, host_ref, host_type):
        """Updates host type for a given host."""
        data = {'hostType': host_type}
        return self.update_host(host_ref, data)

    def set_host_group_for_host(self, host_ref, host_group_ref=utils.NULL_REF):
        """Sets or clears which host group a host is in."""
        data = {'groupId': host_group_ref}
        self.update_host(host_ref, data)

    def list_host_types(self):
        """Lists host types in storage system."""
        path = "/storage-systems/{system-id}/host-types"
        return self._invoke('GET', path)

    def list_snapshot_groups(self):
        """Lists snapshot groups."""
        path = self.RESOURCE_PATHS['snapshot_groups']
        return self._invoke('GET', path)

    def list_snapshot_group(self, object_id):
        """Retrieve given snapshot group from the array."""
        path = self.RESOURCE_PATHS['snapshot_group']
        return self._invoke('GET', path, **{'object-id': object_id})

    def create_snapshot_group(self, label, object_id, storage_pool_id=None,
                              repo_percent=99, warn_thres=99, auto_del_limit=0,
                              full_policy='failbasewrites'):
        """Creates snapshot group on array."""
        path = self.RESOURCE_PATHS['snapshot_groups']
        data = {'baseMappableObjectId': object_id, 'name': label,
                'storagePoolId': storage_pool_id,
                'repositoryPercentage': repo_percent,
                'warningThreshold': warn_thres,
                'autoDeleteLimit': auto_del_limit, 'fullPolicy': full_policy}
        return self._invoke('POST', path, data)

    def update_snapshot_group(self, group_id, label):
        """Modify a snapshot group on the array."""
        path = self.RESOURCE_PATHS['snapshot_group']
        data = {'name': label}
        return self._invoke('POST', path, data, **{'object-id': group_id})

    def delete_snapshot_group(self, object_id):
        """Deletes given snapshot group from array."""
        path = self.RESOURCE_PATHS['snapshot_group']
        return self._invoke('DELETE', path, **{'object-id': object_id})

    def create_snapshot_image(self, group_id):
        """Creates snapshot image in snapshot group."""
        path = self.RESOURCE_PATHS['snapshot_images']
        data = {'groupId': group_id}
        return self._invoke('POST', path, data)

    def delete_snapshot_image(self, object_id):
        """Deletes given snapshot image in snapshot group."""
        path = self.RESOURCE_PATHS['snapshot_image']
        return self._invoke('DELETE', path, **{'object-id': object_id})

    def list_snapshot_image(self, object_id):
        """Retrieve given snapshot image from the array."""
        path = self.RESOURCE_PATHS['snapshot_image']
        return self._invoke('GET', path, **{'object-id': object_id})

    def list_snapshot_images(self):
        """Lists snapshot images."""
        path = self.RESOURCE_PATHS['snapshot_images']
        return self._invoke('GET', path)

    def create_snapshot_volume(self, image_id, label, base_object_id,
                               storage_pool_id=None,
                               repo_percent=99, full_thres=99,
                               view_mode='readOnly'):
        """Creates snapshot volume."""
        path = self.RESOURCE_PATHS['snapshot_volumes']
        data = {'snapshotImageId': image_id, 'fullThreshold': full_thres,
                'storagePoolId': storage_pool_id,
                'name': label, 'viewMode': view_mode,
                'repositoryPercentage': repo_percent,
                'baseMappableObjectId': base_object_id,
                'repositoryPoolId': storage_pool_id}
        return self._invoke('POST', path, data)

    def update_snapshot_volume(self, snap_vol_id, label=None, full_thres=None):
        """Modify an existing snapshot volume."""
        path = self.RESOURCE_PATHS['snapshot_volume']
        data = {'name': label, 'fullThreshold': full_thres}
        return self._invoke('POST', path, data, **{'object-id': snap_vol_id})

    def delete_snapshot_volume(self, object_id):
        """Deletes given snapshot volume."""
        path = self.RESOURCE_PATHS['snapshot_volume']
        return self._invoke('DELETE', path, **{'object-id': object_id})

    def list_snapshot_volumes(self):
        """Lists snapshot volumes/views defined on the array."""
        path = self.RESOURCE_PATHS['snapshot_volumes']
        return self._invoke('GET', path)

    def list_ssc_storage_pools(self):
        """Lists pools and their service quality defined on the array."""
        path = "/storage-systems/{system-id}/ssc/pools"
        return self._invoke('GET', path)

    def get_ssc_storage_pool(self, volume_group_ref):
        """Get storage pool service quality information from the array."""
        path = "/storage-systems/{system-id}/ssc/pools/{object-id}"
        return self._invoke('GET', path, **{'object-id': volume_group_ref})

    def list_storage_pools(self):
        """Lists storage pools in the array."""
        path = "/storage-systems/{system-id}/storage-pools"
        return self._invoke('GET', path)

    def get_storage_pool(self, volume_group_ref):
        """Get storage pool information from the array."""
        path = "/storage-systems/{system-id}/storage-pools/{object-id}"
        return self._invoke('GET', path, **{'object-id': volume_group_ref})

    def list_drives(self):
        """Lists drives in the array."""
        path = "/storage-systems/{system-id}/drives"
        return self._invoke('GET', path)

    def list_storage_systems(self):
        """Lists managed storage systems registered with web service."""
        path = "/storage-systems"
        return self._invoke('GET', path, use_system=False)

    def list_storage_system(self):
        """List current storage system registered with web service."""
        path = "/storage-systems/{system-id}"
        return self._invoke('GET', path)

    def register_storage_system(self, controller_addresses, password=None,
                                wwn=None):
        """Registers storage system with web service."""
        path = "/storage-systems"
        data = {'controllerAddresses': controller_addresses}
        data.setdefault('wwn', wwn if wwn else None)
        data.setdefault('password', password if password else None)
        return self._invoke('POST', path, data, use_system=False)

    def update_stored_system_password(self, password):
        """Update array password stored on web service."""
        path = "/storage-systems/{system-id}"
        data = {'storedPassword': password}
        return self._invoke('POST', path, data)

    def create_volume_copy_job(self, src_id, tgt_id, priority='priority4',
                               tgt_wrt_protected='true'):
        """Creates a volume copy job."""
        path = "/storage-systems/{system-id}/volume-copy-jobs"
        data = {'sourceId': src_id, 'targetId': tgt_id,
                'copyPriority': priority,
                'targetWriteProtected': tgt_wrt_protected}
        return self._invoke('POST', path, data)

    def control_volume_copy_job(self, obj_id, control='start'):
        """Controls a volume copy job."""
        path = ("/storage-systems/{system-id}/volume-copy-jobs-control"
                "/{object-id}?control={String}")
        return self._invoke('PUT', path, **{'object-id': obj_id,
                                            'String': control})

    def list_vol_copy_job(self, object_id):
        """List volume copy job."""
        path = "/storage-systems/{system-id}/volume-copy-jobs/{object-id}"
        return self._invoke('GET', path, **{'object-id': object_id})

    def delete_vol_copy_job(self, object_id):
        """Delete volume copy job."""
        path = "/storage-systems/{system-id}/volume-copy-jobs/{object-id}"
        return self._invoke('DELETE', path, **{'object-id': object_id})

    def add_autosupport_data(self, key, data):
        """Register driver statistics via autosupport log."""
        path = ('/key-values/%s' % key)
        self._invoke('POST', path, json.dumps(data))

    def set_counter(self, key, value):
        path = ('/counters/%s/setCounter?value=%d' % (key, value))
        self._invoke('POST', path)

    def get_asup_info(self):
        """Returns a dictionary of relevant autosupport information.

        Currently returned fields are:
        model -- E-series model name
        serial_numbers -- Serial number for each controller
        firmware_version -- Version of active firmware
        chassis_sn -- Serial number for whole chassis
        """
        asup_info = {}

        controllers = self.list_hardware_inventory().get('controllers')
        if controllers:
            asup_info['model'] = controllers[0].get('modelName', 'unknown')
            serial_numbers = [value['serialNumber'].rstrip()
                              for __, value in enumerate(controllers)]
            serial_numbers.sort()
            for index, value in enumerate(serial_numbers):
                if not value:
                    serial_numbers[index] = 'unknown'
            asup_info['serial_numbers'] = serial_numbers
        else:
            asup_info['model'] = 'unknown'
            asup_info['serial_numbers'] = ['unknown', 'unknown']

        system_info = self.list_storage_system()
        if system_info:
            asup_info['firmware_version'] = system_info['fwVersion']
            asup_info['chassis_sn'] = system_info['chassisSerialNumber']
        else:
            asup_info['firmware_version'] = 'unknown'
            asup_info['chassis_sn'] = 'unknown'

        return asup_info

    def get_eseries_api_info(self, verify=False):
        """Get E-Series API information from the array."""
        api_operating_mode = 'embedded'
        path = 'devmgr/utils/about'
        headers = {'Content-Type': 'application/json',
                   'Accept': 'application/json'}
        url = self._get_resource_url(path, True).replace(
            '/devmgr/v2', '', 1)
        result = self.invoke_service(method='GET', url=url,
                                     headers=headers,
                                     verify=verify)
        about_response_dict = result.json()
        mode_is_proxy = about_response_dict['runningAsProxy']
        if mode_is_proxy:
            api_operating_mode = 'proxy'
        return api_operating_mode, about_response_dict['version']

    def list_backend_store(self, key):
        """Retrieve data by key from the the persistent store on the backend.

        Example response: {"key": "cinder-snapshots", "value": "[]"}

        :param key: the persistent store to retrieve
        :return a json body representing the value of the store,
        or an empty json object
        """
        path = self.RESOURCE_PATHS.get('persistent-store')
        try:
            resp = self._invoke('GET', path, **{'key': key})
        except exception.NetAppDriverException:
            return dict()
        else:
            data = resp['value']
            if data:
                return json.loads(data)
            return dict()

    def save_backend_store(self, key, store_data):
        """Save a json value to the persistent storage on the backend.

        The storage backend provides a small amount of persistent storage
        that we can utilize for storing driver information.

        :param key: The key utilized for storing/retrieving the data
        :param store_data: a python data structure that will be stored as a
        json value
        """
        path = self.RESOURCE_PATHS.get('persistent-stores')
        store_data = json.dumps(store_data, separators=(',', ':'))

        data = {
            'key': key,
            'value': store_data
        }

        self._invoke('POST', path, data)



#
# Copyright 2015 Nexenta Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from cinder import context
from cinder import test
from cinder.volume import configuration as conf
from cinder.volume.drivers.nexenta.nexentaedge import iscsi

NEDGE_URL = 'service/isc/iscsi'
NEDGE_BUCKET = 'c/t/bk'
NEDGE_SERVICE = 'isc'
NEDGE_BLOCKSIZE = 4096
NEDGE_CHUNKSIZE = 16384

MOCK_VOL = {
    'id': 'vol1',
    'name': 'vol1',
    'size': 1
}
MOCK_VOL2 = {
    'id': 'vol2',
    'name': 'vol2',
    'size': 1
}
MOCK_SNAP = {
    'id': 'snap1',
    'name': 'snap1',
    'volume_name': 'vol1'
}
NEW_VOL_SIZE = 2
ISCSI_TARGET_NAME = 'iscsi_target_name'
ISCSI_TARGET_STATUS = 'Target 1: ' + ISCSI_TARGET_NAME


class TestNexentaEdgeISCSIDriver(test.TestCase):

    def setUp(self):
        super(TestNexentaEdgeISCSIDriver, self).setUp()
        self.cfg = mock.Mock(spec=conf.Configuration)
        self.cfg.nexenta_client_address = '0.0.0.0'
        self.cfg.nexenta_rest_address = '0.0.0.0'
        self.cfg.nexenta_rest_port = 8080
        self.cfg.nexenta_rest_protocol = 'http'
        self.cfg.nexenta_iscsi_target_portal_port = 3260
        self.cfg.nexenta_rest_user = 'admin'
        self.cfg.nexenta_rest_password = 'admin'
        self.cfg.nexenta_lun_container = NEDGE_BUCKET
        self.cfg.nexenta_iscsi_service = NEDGE_SERVICE
        self.cfg.nexenta_blocksize = NEDGE_BLOCKSIZE
        self.cfg.nexenta_chunksize = NEDGE_CHUNKSIZE

        mock_exec = mock.Mock()
        mock_exec.return_value = ('', '')
        self.driver = iscsi.NexentaEdgeISCSIDriver(execute=mock_exec,
                                                   configuration=self.cfg)
        self.api_patcher = mock.patch('cinder.volume.drivers.nexenta.'
                                      'nexentaedge.jsonrpc.'
                                      'NexentaEdgeJSONProxy.__call__')
        self.mock_api = self.api_patcher.start()

        self.mock_api.return_value = {
            'data': {'value': ISCSI_TARGET_STATUS}
        }
        self.driver.do_setup(context.get_admin_context())

        self.addCleanup(self.api_patcher.stop)

    def test_check_do_setup(self):
        self.assertEqual(ISCSI_TARGET_NAME, self.driver.target_name)

    def test_create_volume(self):
        self.driver.create_volume(MOCK_VOL)
        self.mock_api.assert_called_with(NEDGE_URL, {
            'objectPath': NEDGE_BUCKET + '/' + MOCK_VOL['id'],
            'volSizeMB': MOCK_VOL['size'] * 1024,
            'blockSize': NEDGE_BLOCKSIZE,
            'chunkSize': NEDGE_CHUNKSIZE
        })

    def test_create_volume_fail(self):
        self.mock_api.side_effect = RuntimeError
        self.assertRaises(RuntimeError, self.driver.create_volume, MOCK_VOL)

    def test_delete_volume(self):
        self.driver.delete_volume(MOCK_VOL)
        self.mock_api.assert_called_with(NEDGE_URL, {
            'objectPath': NEDGE_BUCKET + '/' + MOCK_VOL['id']
        })

    def test_delete_volume_fail(self):
        self.mock_api.side_effect = RuntimeError
        self.assertRaises(RuntimeError, self.driver.delete_volume, MOCK_VOL)

    def test_extend_volume(self):
        self.driver.extend_volume(MOCK_VOL, NEW_VOL_SIZE)
        self.mock_api.assert_called_with(NEDGE_URL + '/resize', {
            'objectPath': NEDGE_BUCKET + '/' + MOCK_VOL['id'],
            'newSizeMB': NEW_VOL_SIZE * 1024
        })

    def test_extend_volume_fail(self):
        self.mock_api.side_effect = RuntimeError
        self.assertRaises(RuntimeError, self.driver.extend_volume,
                          MOCK_VOL, NEW_VOL_SIZE)

    def test_create_snapshot(self):
        self.driver.create_snapshot(MOCK_SNAP)
        self.mock_api.assert_called_with(NEDGE_URL + '/snapshot', {
            'objectPath': NEDGE_BUCKET + '/' + MOCK_VOL['id'],
            'snapName': MOCK_SNAP['id']
        })

    def test_create_snapshot_fail(self):
        self.mock_api.side_effect = RuntimeError
        self.assertRaises(RuntimeError, self.driver.create_snapshot, MOCK_SNAP)

    def test_delete_snapshot(self):
        self.driver.delete_snapshot(MOCK_SNAP)
        self.mock_api.assert_called_with(NEDGE_URL + '/snapshot', {
            'objectPath': NEDGE_BUCKET + '/' + MOCK_VOL['id'],
            'snapName': MOCK_SNAP['id']
        })

    def test_delete_snapshot_fail(self):
        self.mock_api.side_effect = RuntimeError
        self.assertRaises(RuntimeError, self.driver.delete_snapshot, MOCK_SNAP)

    def test_create_volume_from_snapshot(self):
        self.driver.create_volume_from_snapshot(MOCK_VOL2, MOCK_SNAP)
        self.mock_api.assert_called_with(NEDGE_URL + '/snapshot/clone', {
            'objectPath': NEDGE_BUCKET + '/' + MOCK_SNAP['volume_name'],
            'clonePath': NEDGE_BUCKET + '/' + MOCK_VOL2['id'],
            'snapName': MOCK_SNAP['id']
        })

    def test_create_volume_from_snapshot_fail(self):
        self.mock_api.side_effect = RuntimeError
        self.assertRaises(RuntimeError,
                          self.driver.create_volume_from_snapshot,
                          MOCK_VOL2, MOCK_SNAP)

    def test_create_cloned_volume(self):
        self.driver.create_cloned_volume(MOCK_VOL2, MOCK_VOL)
        self.mock_api.assert_called_with(NEDGE_URL, {
            'objectPath': NEDGE_BUCKET + '/' + MOCK_VOL2['id'],
            'volSizeMB': MOCK_VOL2['size'] * 1024,
            'blockSize': NEDGE_BLOCKSIZE,
            'chunkSize': NEDGE_CHUNKSIZE
        })

    def test_create_cloned_volume_fail(self):
        self.mock_api.side_effect = RuntimeError
        self.assertRaises(RuntimeError, self.driver.create_cloned_volume,
                          MOCK_VOL2, MOCK_VOL)
